# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# mypy: disable-error-code="no-untyped-def"
"""Pytest fixtures to manage and access a local SciCat backend."""

import logging
from contextlib import contextmanager
from pathlib import Path
from typing import Generator, Optional, Type, Union

import pytest

from ..._internal import docker
from ..._internal.file_counter import FileCounter, NullCounter
from ...client import Client
from .._pytest_helpers import init_pytest_work_dir
from ..client import FakeClient
from . import seed
from ._backend import configure, wait_until_backend_is_live
from ._pytest_helpers import backend_enabled, skip_if_not_backend
from .config import SciCatAccess, local_access


@pytest.fixture(scope="session")
def scicat_access() -> SciCatAccess:
    """Fixture that returns SciCat access parameters.

    Returns
    -------
    :
        A URL and user to connect to the testing backend.
        The user has access to all initial datasets in the database and permissions
        to create new datasets.
    """
    return local_access("user1")


@pytest.fixture()
def fake_client(scicat_access: SciCatAccess) -> FakeClient:
    """Fixture that returns a fake client.

    The client is seeded from the same data as the backend.
    But only if the backend has been requested.
    So any test that uses ``fake_client`` explicitly and requires the seed,
    must also request the ``require_scicat_backend`` fixture.

    Returns
    -------
    :
        A fake client that potentially contains the seed data.
        See :mod:`scitacean.testing.backend.seed`.
    """
    client = FakeClient.from_credentials(
        url=scicat_access.url,
        **scicat_access.user.credentials,  # type: ignore[arg-type]
    )
    client.datasets.update(
        {ds.pid: ds for ds in seed.INITIAL_DATASETS.values()}  # type: ignore[misc]
    )
    client.orig_datablocks.update(
        {
            dbs[0].datasetId: dbs  # type: ignore[misc]
            for dbs in seed.INITIAL_ORIG_DATABLOCKS.values()
        }
    )
    client.attachments.update(
        {
            a[0].datasetId: a  # type: ignore[misc]
            for a in seed.INITIAL_ATTACHMENTS.values()
        }
    )
    return client


@pytest.fixture()
def real_client(scicat_access: SciCatAccess, scicat_backend: bool) -> Optional[Client]:
    """Fixture that returns a real client if backend tests are enabled.

    Returns
    -------
    :
        If backend tests are enabled, a real client that is connected to
        the testing backend.
        Otherwise, it returns None.
    """
    if not scicat_backend:
        return None
    return Client.from_credentials(
        url=scicat_access.url,
        **scicat_access.user.credentials,  # type: ignore[arg-type]
    )


@pytest.fixture(params=["real", "fake"])
def client(request, scicat_backend) -> Union[Client, FakeClient]:
    """Fixture that returns a real and a fake client.

    Using this fixture makes tests run twice, once with a real client
    and once with a fake client.
    If backend tests are disabled, the test with a real client gets skipped.

    Returns
    -------
    :
        A real client and a fake client.
    """
    if request.param == "fake":
        return request.getfixturevalue("fake_client")  # type: ignore[no-any-return]
    skip_if_not_backend(request)
    return request.getfixturevalue("real_client")  # type: ignore[no-any-return]


@pytest.fixture()
def require_scicat_backend(request, scicat_backend) -> None:
    """Fixture to declare that a test needs a local scicat backend.

    Like :func:`scitacean.testing.backend.scicat_backend`
    but this skips the test if backend tests are disabled.
    """
    skip_if_not_backend(request)


@pytest.fixture(scope="session")
def scicat_backend(request, tmp_path_factory, scicat_access):
    """Fixture to declare that a test needs a local scicat backend.

    If backend tests are enabled, this fixture spins up the backend in a docker
    container the first time a test requests it.
    The backend will be stopped and removed at the end of the test session.
    :mod:`scitacean.testing.backend.seed` is used to initialize the database.

    If backend tests are disabled, this fixture prepares the seed for use with
    :func:`scitacean.testing.backend.fixtures.fake_client`.

    Returns
    -------
    :
        True if backend tests are enabled and False otherwise.
    """
    target_dir, counter = init_pytest_work_dir(
        request, tmp_path_factory, name="scitacean-scicat-backend"
    )

    if not backend_enabled(request):
        with _prepare_without_backend(
            scicat_access=scicat_access, counter=counter, target_dir=target_dir
        ):
            yield False
    else:
        with _prepare_with_backend(
            scicat_access=scicat_access, target_dir=target_dir, counter=counter
        ):
            yield True


@contextmanager
def _prepare_without_backend(
    scicat_access: SciCatAccess,
    counter: Union[FileCounter, NullCounter],
    target_dir: Path,
) -> Generator[None, None, None]:
    with counter.increment() as count:
        if count == 1:
            _seed_database(
                FakeClient, scicat_access=scicat_access, target_dir=target_dir
            )
        else:
            seed.seed_worker(target_dir)
    yield


@contextmanager
def _prepare_with_backend(
    scicat_access: SciCatAccess,
    counter: Union[FileCounter, NullCounter],
    target_dir: Path,
) -> Generator[None, None, None]:
    try:
        with counter.increment() as count:
            if count == 1:
                _backend_docker_up(target_dir)
                _seed_database(
                    Client, scicat_access=scicat_access, target_dir=target_dir
                )
            elif not _backend_is_running():
                raise RuntimeError("Expected backend to be running")
            else:
                seed.seed_worker(target_dir)
        yield
    finally:
        with counter.decrement() as count:
            if count == 0:
                _backend_docker_down(target_dir)


def _seed_database(
    client_class: Union[Type[Client], Type[FakeClient]],
    scicat_access: SciCatAccess,
    target_dir: Path,
) -> None:
    client = client_class.from_credentials(
        url=scicat_access.url,
        **scicat_access.user.credentials,  # type: ignore[arg-type]
    )
    seed.seed_database(client=client, scicat_access=scicat_access)
    seed.save_seed(target_dir)


def _backend_docker_up(target_dir: Path) -> None:
    if _backend_is_running():
        raise RuntimeError("SciCat docker container is already running")
    docker_compose_file = target_dir / "docker-compose.yaml"
    log = logging.getLogger("scitacean.testing")
    log.info(
        "Starting docker container with SciCat backend from %s", docker_compose_file
    )
    configure(docker_compose_file)
    docker.docker_compose_up(docker_compose_file)
    log.info("Waiting for SciCat docker to become accessible")
    wait_until_backend_is_live(max_time=60, n_tries=40)
    log.info("Successfully connected to SciCat backend")


def _backend_docker_down(target_dir: Path) -> None:
    # Check if container is running because the fixture can call this function
    # if there was an exception in _backend_docker_up.
    # In that case, there is nothing to tear down.
    if _backend_is_running():
        docker_compose_file = target_dir / "docker-compose.yaml"
        log = logging.getLogger("scitacean.testing")
        log.info(
            "Stopping docker container with SciCat backend from %s", docker_compose_file
        )
        docker.docker_compose_down(docker_compose_file)


def _backend_is_running() -> bool:
    return docker.container_is_running("scitacean-test-scicat")


__all__ = [
    "scicat_access",
    "scicat_backend",
    "real_client",
    "fake_client",
    "client",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Utilities for running tests against a locally deployed SciCat backend.

This subpackage provides tools for using a real SciCat backend running in docker
containers on the local machine.
It, therefore, requires docker to be installed and running.
The package is in particular intended to be used with
`pytest <https://docs.pytest.org>`_ and the fixtures provided by the
:mod:`scitacean.testing.backend.fixtures` submodule.
But it is also possible to set up tests in a different way by using the functions
provided here.

Note
----
Although this package attempts to clean up after itself, it is possible that it
fails to tear down the backend completely if tests are aborted in an unusual way
(pressing Ctrl-C is fine).
If this happens, a running docker container may be left behind.
It needs to be stopped manually.
Failing this, future tests may freeze indefinitely during start up.

Attention
---------
The fixtures support `pytest-xdist <https://pytest-xdist.readthedocs.io/en/latest/>`_
but only if all workers run on the local machine (the default).

It may still happen that tests fail due to the complexity of synchronizing start up
and shut down of the backend between workers.

See Also
--------
`Testing <../../user-guide/testing.ipynb>`_ user guide.

Examples
--------
The following pytest test will download the "raw" dataset that is included in the seed.
It runs both with a real and a fake client.

.. code-block:: python

    from scitacean.testing.backend import seed

    def test_download(client):
        ds = client.get_dataset(seed.INITIAL_DATASETS["raw"].pid)
        # assert something

If a test requires a backend but wants to construct a client manually, use

.. code-block:: python

    def test_manual_client(require_scicat_backend, scicat_access):
        client = Client.from_credentials(
            url=scicat_access.url,
            **scicat_access.user.credentials
        )
        # assert something

.. rubric:: Functions

.. autosummary::
   :toctree: ../functions

   add_pytest_option
   backend_enabled
   can_connect
   configure
   skip_if_not_backend
   start_backend
   stop_backend
   wait_until_backend_is_live
"""

from . import config, seed
from ._backend import (
    can_connect,
    configure,
    start_backend,
    stop_backend,
    wait_until_backend_is_live,
)
from ._pytest_helpers import add_pytest_option, backend_enabled, skip_if_not_backend

__all__ = [
    "add_pytest_option",
    "backend_enabled",
    "can_connect",
    "config",
    "configure",
    "seed",
    "skip_if_not_backend",
    "start_backend",
    "stop_backend",
    "wait_until_backend_is_live",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""HTML representations for Jupyter."""
from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional

from ._attachment_html import attachment_html_repr
from ._dataset_html import dataset_html_repr

if TYPE_CHECKING:
    from ..model import BaseUserModel


@lru_cache(maxsize=1)
def _user_model_reprs() -> Dict[type, Callable[[Any], str]]:
    from ..model import Attachment

    return {Attachment: attachment_html_repr}


def user_model_html_repr(user_model: BaseUserModel) -> Optional[str]:
    """HTML representation of a user model f implemented."""
    if (repr_fn := _user_model_reprs().get(type(user_model))) is not None:
        return repr_fn(user_model)
    return None


__all__ = ["dataset_html_repr", "user_model_html_repr"]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

from typing import Optional

import pytest

_COMMAND_LINE_OPTION: Optional[str] = None


def add_pytest_option(parser: pytest.Parser, option: str = "--ssh-tests") -> None:
    """Add a command-line option to pytest to toggle SSH tests.

    Parameters
    ----------
    parser:
        Pytest's command-line argument parser.
    option:
        Name of the command-line option.
    """
    parser.addoption(
        option,
        action="store_true",
        default=False,
        help="Select whether to run tests with an SSH fileserver",
    )
    global _COMMAND_LINE_OPTION
    _COMMAND_LINE_OPTION = option


def skip_if_not_ssh(request: pytest.FixtureRequest) -> None:
    """Mark the current test to be skipped if SSH tests are disabled."""
    if not ssh_enabled(request):
        pytest.skip(
            "Tests against an SSH file server are disabled, "
            f"use {_COMMAND_LINE_OPTION} to enable them"
        )


def ssh_enabled(request: pytest.FixtureRequest) -> bool:
    """Return True if SSH tests are enabled."""
    return _COMMAND_LINE_OPTION is not None and bool(
        request.config.getoption(_COMMAND_LINE_OPTION)
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

"""Types and functions to implement models for communication with SciCat."""
from __future__ import annotations

import dataclasses
from datetime import datetime
from typing import (
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

import pydantic
from dateutil.parser import parse as parse_datetime

from ._internal.orcid import is_valid_orcid
from ._internal.pydantic_compat import is_pydantic_v1
from .filesystem import RemotePath
from .logging import get_logger
from .pid import PID
from .thumbnail import Thumbnail

try:
    # Python 3.11+
    from enum import StrEnum

    class DatasetType(StrEnum):
        """Type of Dataset."""

        RAW = "raw"
        DERIVED = "derived"

    del StrEnum

except ImportError:
    from enum import Enum

    class DatasetType(str, Enum):  # type: ignore[no-redef]
        """Enum representing the type of datasets."""

        RAW = "raw"
        DERIVED = "derived"

    del Enum


class BaseModel(pydantic.BaseModel):
    """Base class for Pydantic models for communication with SciCat."""

    if is_pydantic_v1():

        class Config:
            extra = pydantic.Extra.forbid
            json_encoders = {  # noqa: RUF012
                PID: lambda v: str(v),
                RemotePath: lambda v: v.posix,
                Thumbnail: lambda v: v.serialize(),
            }

    else:
        model_config = pydantic.ConfigDict(
            extra="forbid",
        )

    _user_mask: ClassVar[Tuple[str, ...]]
    _masked_fields: ClassVar[Optional[Tuple[str, ...]]] = None

    # Some schemas contain fields that we don't want to use in Scitacean.
    # Normally, omitting them from the model would result in an error when
    # building a model from the JSON returned by SciCat.
    # The following subclass hook allows models to mark fields as masked.
    # Those will be silently dropped by __init__.
    # Note also the comment for _IGNORED_KWARGS below.
    def __init_subclass__(
        cls, /, masked: Optional[Iterable[str]] = None, **kwargs: Any
    ) -> None:
        super().__init_subclass__(**kwargs)
        cls._user_mask = tuple(masked) if masked is not None else ()

    def __init__(self, **kwargs: Any) -> None:
        self._delete_ignored_args(kwargs)
        super().__init__(**kwargs)

    def _delete_ignored_args(self, args: Dict[str, Any]) -> None:
        if self._masked_fields is None:
            self._init_mask(self)
        for key in self._masked_fields:  # type: ignore[union-attr]
            args.pop(key, None)

    # Initializing the mask requires the field names which
    # are only available on instances.
    # So initialization needs to be deferred until the first instantiation of the model.
    # The mask is cached afterward.
    @classmethod
    def _init_mask(cls: Type[ModelType], instance: ModelType) -> None:
        def get_name(name: str, field: Any) -> Any:
            return field.alias if field.alias is not None else name

        field_names = {
            get_name(name, field) for name, field in instance.get_model_fields().items()
        }
        default_mask = tuple(key for key in _IGNORED_KWARGS if key not in field_names)
        cls._masked_fields = cls._user_mask + default_mask

    @classmethod
    def user_model_type(cls) -> Optional[Type[BaseUserModel]]:
        """Return the user model type for this model.

        Returns ``None`` if there is no user model, e.g., for ``Dataset``
        where there is a custom class instead of a plain model.
        """
        return None

    @classmethod
    def upload_model_type(cls) -> Optional[Type[BaseModel]]:
        """Return the upload model type for this model.

        Returns ``None`` if the model cannot be uploaded or this is an upload model.
        """
        return None

    @classmethod
    def download_model_type(cls) -> Optional[Type[BaseModel]]:
        """Return the download model type for this model.

        Returns ``None`` if this is a download model.
        """
        return None

    if is_pydantic_v1():

        @classmethod
        def get_model_fields(cls) -> Dict[str, Any]:
            return cls.__fields__  # type: ignore[return-value]

        def model_dump(self, *args: Any, **kwargs: Any) -> Dict[str, Any]:
            return self.dict(*args, **kwargs)

        def model_dump_json(self, *args: Any, **kwargs: Any) -> str:
            return self.json(*args, **kwargs)

        @classmethod
        def model_construct(
            cls: Type[ModelType], *args: Any, **kwargs: Any
        ) -> ModelType:
            return cls.construct(*args, **kwargs)

        @classmethod
        def model_rebuild(cls, *args: Any, **kwargs: Any) -> Optional[bool]:
            return cls.update_forward_refs(*args, **kwargs)

    else:

        @classmethod
        def get_model_fields(cls) -> Dict[str, pydantic.fields.FieldInfo]:
            return cls.model_fields


@dataclasses.dataclass
class BaseUserModel:
    """Base class for user models.

    Child classes must be dataclasses.
    """

    @classmethod
    def _download_model_dict(cls, download_model: Any) -> Dict[str, Any]:
        return {
            field.name: getattr(
                download_model, _model_field_name_of(cls.__name__, field.name)
            )
            for field in dataclasses.fields(cls)
        }

    def _upload_model_dict(self) -> Dict[str, Any]:
        _check_ready_for_upload(self)
        return {
            _model_field_name_of(self.__class__.__name__, field.name): getattr(
                self, field.name
            )
            for field in dataclasses.fields(self)
            if not field.name.startswith("_")
        }

    @classmethod
    def from_download_model(cls, download_model: Any) -> BaseUserModel:
        raise NotImplementedError("Function does not exist for BaseUserModel")

    def make_upload_model(self) -> BaseModel:
        raise NotImplementedError("Function does not exist for BaseUserModel")

    @classmethod
    def upload_model_type(cls) -> Optional[Type[BaseModel]]:
        """Return the upload model type for this user model.

        Returns ``None`` if the model cannot be uploaded.
        """
        return None

    @classmethod
    def download_model_type(cls) -> Type[BaseModel]:
        """Return the download model type for this user model."""
        # There is no sensible default value here as there always exists a download
        # model.
        # All child classes must implement this function.
        raise NotImplementedError("Function does not exist for BaseUserModel")

    def _repr_html_(self) -> Optional[str]:
        """Return an HTML representation of the model if possible."""
        from ._html_repr import user_model_html_repr

        return user_model_html_repr(self)


def construct(
    model: Type[PydanticModelType],
    *,
    _strict_validation: bool = True,
    _quiet: bool = False,
    **fields: Any,
) -> PydanticModelType:
    """Instantiate a SciCat model.

    Warning
    -------
    If the model is created without validation, no fields will be converted
    to their proper type but will simply be whatever arguments are passed.
    See ``model_construct`` or :class:`pydantic.BaseModel` for more information.

    A warning will be emitted in this case.

    Parameters
    ----------
    model:
        Class of the model to create.
    _strict_validation:
        If ``True``, the model must pass validation.
        If ``False``, a model is still returned if validation fails.
    _quiet:
        If ``False``, logs a warning on validation failure.
    fields:
        Field values to pass to the model initializer.

    Returns
    -------
    :
        An initialized model.
    """
    try:
        return model(**fields)
    except pydantic.ValidationError as e:
        if _strict_validation:
            raise
        if not _quiet:
            get_logger().warning(
                "Validation of metadata failed: %s\n"
                "The returned object may be incomplete or broken. "
                "In particular, some fields may not have the correct type",
                str(e),
            )
        return model.model_construct(**fields)


def validate_datetime(value: Optional[Union[str, datetime]]) -> Optional[datetime]:
    """Convert strings to datetimes.

    This uses dateutil.parser.parse instead of Pydantic's builtin parser in order to
    produce results that are consistent with user inputs.
    Pydantic uses a custom type for timezones which is not fully compatible with
    dateutil's.
    """
    if not isinstance(value, str):
        return value
    return parse_datetime(value)


def validate_drop(_: Any) -> None:
    """Return ``None``."""
    return None


def validate_emails(value: Optional[str]) -> Optional[str]:
    if value is None:
        return value
    return ";".join(pydantic.validate_email(item)[1] for item in value.split(";"))


def validate_orcids(value: Optional[str]) -> Optional[str]:
    if value is None:
        return value
    try:
        if is_valid_orcid(value):
            return value
    except (RuntimeError, ValueError, TypeError):
        pass
    raise ValueError(
        "value is not a valid ORCID, "
        "note that ORCIDs must be prefixed with 'https://orcid.org'."
    )


@overload
def convert_download_to_user_model(download_model: None) -> None:
    ...


@overload
def convert_download_to_user_model(download_model: BaseModel) -> BaseUserModel:
    ...


@overload
def convert_download_to_user_model(
    download_model: Iterable[BaseModel],
) -> List[BaseUserModel]:
    ...


def convert_download_to_user_model(
    download_model: Optional[Union[BaseModel, Iterable[BaseModel]]]
) -> Optional[Union[BaseUserModel, List[BaseUserModel]]]:
    """Construct user models from download models."""
    if download_model is None:
        return download_model
    if isinstance(download_model, BaseModel):
        if (user_type := download_model.user_model_type()) is None:
            raise TypeError("Cannot convert to user model in this way.")
        return user_type.from_download_model(download_model)
    return list(map(convert_download_to_user_model, download_model))


@overload
def convert_user_to_upload_model(user_model: None) -> None:
    ...


@overload
def convert_user_to_upload_model(user_model: BaseUserModel) -> BaseModel:
    ...


@overload
def convert_user_to_upload_model(
    user_model: Iterable[BaseUserModel],
) -> List[BaseModel]:
    ...


def convert_user_to_upload_model(
    user_model: Optional[Union[BaseUserModel, Iterable[BaseUserModel]]]
) -> Optional[Union[BaseModel, List[BaseModel]]]:
    """Construct upload models from user models."""
    if user_model is None:
        return None
    if isinstance(user_model, BaseUserModel):
        return user_model.make_upload_model()
    return list(map(convert_user_to_upload_model, user_model))


def _model_field_name_of(cls_name: str, name: str) -> str:
    """Convert a user model field name to a SciCat model field name.

    Converts snake_case to camelCase and strips leading underscores.
    E.g.,
    `proposal_id` -> `proposalId`,
    `_created_at` -> `createdAt`,
    `_History__id` -> `id`.
    """
    name = name.lstrip("_")
    if name.startswith(cls_name):
        # Field names with two leading underscores are prefixed with `_cls_name`
        # by dataclasses, e.g. `__id` -> `_History__id`.
        # Strip this prefix plus underscores.
        name = name[len(cls_name) + 2 :]
    first, *remainder = name.split("_")
    return first + "".join(word.capitalize() for word in remainder)


def _check_ready_for_upload(user_model: BaseUserModel) -> None:
    download_only = {
        field.name: value
        for field in dataclasses.fields(user_model)
        if (value := getattr(user_model, field.name)) is not None
        and field.name.startswith("_")
    }
    if download_only:
        raise ValueError(
            f"These fields of {user_model.__class__.__name__} "
            "must not be set during upload:\n"
            + "\n".join(f"  {name} = {value}" for name, value in download_only.items())
        )


# These fields are auto generated by MongoDB/Mongoose,
# but we don't want them in Scitacean.
# Drop them in the constructor if they are not in the model explicitly.
# - id is a unique ID used by backend v3.*.
#   Used to be hidden in the API for datasets in favor of pid.
#   (Should be the same as pid where applicable.)
# - _id is MongoDB's unique ID for the entry.
#   (Should be the same as id/pid where applicable.)
# - _v or __v is some version added by Mongoose.
_IGNORED_KWARGS = ("id", "_id", "_v", "__v")

ModelType = TypeVar("ModelType", bound=BaseModel)
PydanticModelType = TypeVar("PydanticModelType", bound=pydantic.BaseModel)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

import hypothesis
import pytest

from scitacean.testing.backend import add_pytest_option as add_backend_option
from scitacean.testing.sftp import add_pytest_option as add_sftp_option
from scitacean.testing.ssh import add_pytest_option as add_ssh_option

pytest_plugins = (
    "scitacean.testing.backend.fixtures",
    "scitacean.testing.sftp.fixtures",
    "scitacean.testing.ssh.fixtures",
)


# The datasets strategy requires a large amount of memory and time.
# This is not good but hard to avoid.
# So simply disable health checks and accept that tests are slow.
hypothesis.settings.register_profile(
    "scitacean",
    suppress_health_check=[
        hypothesis.HealthCheck.data_too_large,
        hypothesis.HealthCheck.too_slow,
    ],
)


def pytest_addoption(parser: pytest.Parser) -> None:
    add_backend_option(parser)
    add_sftp_option(parser)
    add_ssh_option(parser)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Load model specifications."""
import dataclasses
import re
import sys
from copy import deepcopy
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union

import yaml

from .schema import Schema, SchemaField, load_schemas


@dataclasses.dataclass
class _UpDownSchemas:
    download: Schema
    upload: Optional[Schema]


@dataclasses.dataclass
class _DatasetSchemas:
    download: Schema
    upload_derived: Schema
    upload_raw: Schema


@dataclasses.dataclass
class SpecField:
    name: str
    scicat_name: str
    description: str
    type: str
    required: bool  # Required in upload.
    default: Optional[str] = None
    upload: bool = False
    download: bool = False
    validation: Optional[str] = None

    def full_type_for(self, kind: Literal["download", "upload", "user"]) -> str:
        return (
            self.type_for(kind)
            if self.required and kind != "download"
            else f"Optional[{self.type_for(kind)}]"
        )

    def type_for(self, kind: Literal["download", "upload", "user"]) -> str:
        """Translate SciCat schema/DTO names into Scitacean model names."""
        if kind == "upload":
            prefix = "Upload"
        elif kind == "download":
            prefix = "Download"
        else:
            prefix = ""

        for spec_name in _SCHEMA_GROUPS:
            if spec_name in self.type:
                return re.sub(str(spec_name), f"{prefix}{spec_name}", self.type)
        return self.type


@dataclasses.dataclass
class Spec:
    name: str
    download_name: str = dataclasses.field(init=False)
    upload_name: Optional[str] = dataclasses.field(init=False)
    fields: Dict[str, SpecField]
    masked_fields_download: Dict[str, SpecField] = dataclasses.field(init=False)
    masked_fields_upload: Dict[str, SpecField] = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        if _SCHEMA_GROUPS.get(self.name, (None, None))[0]:
            self.upload_name = f"Upload{self.name}"
        else:
            self.upload_name = None
        self.download_name = f"Download{self.name}"
        self.masked_fields_download = {}
        self.masked_fields_upload = {}

    def fields_for(
        self, kind: Literal["download", "upload", "user"]
    ) -> List[SpecField]:
        return sorted(
            sorted(
                filter(
                    lambda field: (kind == "download" and field.download)
                    or (kind == "upload" and field.upload)
                    or (kind == "user"),
                    self.fields.values(),
                ),
                key=lambda field: field.name,
            ),
            key=lambda field: not field.required,
        )


@dataclasses.dataclass
class DatasetFieldConversion:
    func: str
    arg_type: str


@dataclasses.dataclass
class DatasetField(SpecField):
    conversion: Optional[DatasetFieldConversion] = None
    default: Optional[Any] = None
    manual: bool = False

    # These are only used for upload models.
    # For downloads, all fields should be considered as used.
    used_by_derived: bool = False
    used_by_raw: bool = False


@dataclasses.dataclass
class DatasetSpec(Spec):
    download_name: str = dataclasses.field(default="DownloadDataset", init=False)
    upload_name: Optional[str] = dataclasses.field(default="UploadDataset", init=False)
    fields: Dict[str, DatasetField]

    def dset_fields_for(
        self,
        kind: Literal["download", "upload", "user"],
        dset_type: Literal["derived", "raw"],
    ) -> List[DatasetField]:
        return list(
            filter(
                lambda field: field.used_by_derived
                if dset_type == "derived"
                else field.used_by_raw,
                self.fields_for(kind),
            )
        )

    def user_dset_fields(self, manual: Optional[bool] = None) -> List[DatasetField]:
        if manual is None:
            return list(self.fields.values())
        return [field for field in self.fields.values() if field.manual == manual]


_SCHEMA_GROUPS = {
    "Attachment": ("CreateAttachmentDto", "Attachment"),
    "OrigDatablock": ("CreateOrigDatablockDto", "OrigDatablock"),
    "Datablock": ("CreateDatasetDatablockDto", "Datablock"),
    "Lifecycle": (None, "LifecycleClass"),
    "Technique": ("TechniqueClass", "TechniqueClass"),
    "Relationship": ("RelationshipClass", "RelationshipClass"),
    "History": (None, "HistoryClass"),
    "DataFile": ("DataFile", "DataFile"),
    "Instrument": (None, "Instrument"),
    "Sample": ("CreateSampleDto", "SampleClass"),
}


def _collect_schemas(
    schemas: Dict[str, Schema]
) -> Dict[str, Union[_UpDownSchemas, _DatasetSchemas]]:
    return {
        "Dataset": _DatasetSchemas(
            upload_derived=schemas["CreateDerivedDatasetDto"],
            upload_raw=schemas["CreateRawDatasetDto"],
            download=schemas["DatasetClass"],
        ),
        **{
            name: _UpDownSchemas(
                download=schemas[down_name],
                upload=schemas[up_name] if up_name else None,
            )
            for name, (up_name, down_name) in _SCHEMA_GROUPS.items()
        },
    }


def _camel_case_to_snake_case(string: str) -> str:
    """Convert a string from camelCase to snake_case."""

    def repl(match):
        return "_" + match[1].lower()

    return re.sub(r"([A-Z])", repl, string)


def _get_common_field_attr(
    name: str, allow_mismatch: bool = False, **fields: SchemaField
) -> Any:
    source_key = None
    val = None
    for key, field in fields.items():
        if field is None:
            continue
        if val is None:
            source_key = key
            val = getattr(field, name)
        else:
            if not allow_mismatch and val != getattr(field, name):
                sys.stderr.write(
                    f"Mismatch in field {name}:\n"
                    f" {source_key}:\n {val}\n {key}:\n {getattr(field, name)}\n"
                )
    return val


def _merge_field(
    name: str, download: Optional[SchemaField], upload: Optional[SchemaField]
) -> SpecField:
    fields = {"download": download, "upload": upload}
    return SpecField(
        name=_camel_case_to_snake_case(_get_common_field_attr("name", **fields)),
        scicat_name=name,
        description=_get_common_field_attr(
            "description", allow_mismatch=True, **fields
        ),
        type=_get_common_field_attr("type", **fields),
        required=upload is not None and upload.required,
        download=download is not None,
        upload=upload is not None,
    )


def _build_spec(name: str, schemas: _UpDownSchemas) -> Spec:
    field_names = set(schemas.download.fields.keys())
    if schemas.upload is not None:
        field_names |= set(schemas.upload.fields.keys())
    fields = {
        name: _merge_field(
            name,
            download=schemas.download.fields.get(name),
            upload=schemas.upload.fields.get(name)
            if schemas.upload is not None
            else None,
        )
        for name in field_names
    }
    return Spec(
        name=name,
        fields=fields,
    )


def _merge_dataset_field(
    name: str,
    download: Optional[SchemaField],
    raw_upload: Optional[SchemaField],
    derived_upload: Optional[SchemaField],
) -> DatasetField:
    fields = {
        "download": download,
        "raw_upload": raw_upload,
        "derived_upload": derived_upload,
    }
    required = (raw_upload is not None and raw_upload.required) or (
        derived_upload is not None and derived_upload.required
    )
    return DatasetField(
        name=_camel_case_to_snake_case(_get_common_field_attr("name", **fields)),
        scicat_name=name,
        description=_get_common_field_attr(
            "description", allow_mismatch=True, **fields
        ),
        type=_get_common_field_attr("type", **fields),
        required=required,
        download=download is not None,
        upload=raw_upload is not None or derived_upload is not None,
        used_by_derived=derived_upload is not None,
        used_by_raw=raw_upload is not None,
    )


def _build_dataset_spec(name: str, schemas: _DatasetSchemas) -> DatasetSpec:
    field_names = (
        schemas.download.fields.keys()
        | schemas.upload_raw.fields.keys()
        | schemas.upload_derived.fields.keys()
    )
    fields = {
        name: _merge_dataset_field(
            name,
            download=schemas.download.fields.get(name),
            raw_upload=schemas.upload_raw.fields.get(name),
            derived_upload=schemas.upload_derived.fields.get(name),
        )
        for name in field_names
    }
    return DatasetSpec(
        name=name,
        fields=fields,
    )


@lru_cache
def _masked_fields() -> Dict[str, List[str]]:
    with open(Path(__file__).resolve().parent / "masked-fields.yml", "r") as f:
        return yaml.safe_load(f)


def _mask_fields(spec: Spec) -> Spec:
    spec = deepcopy(spec)
    masked = _masked_fields()
    for field_name in masked.get(spec.name, []):
        if isinstance(field_name, dict):
            ((field_name, updown),) = field_name.items()
            if updown == "upload":
                spec.fields[field_name].upload = False
                spec.masked_fields_upload[field_name] = spec.fields[field_name]
            elif updown == "download":
                spec.masked_fields_download[field_name] = spec.fields[field_name]
            else:
                raise ValueError("Invalid mask value")
        else:
            spec.masked_fields_upload[field_name] = spec.fields[field_name]
            spec.masked_fields_download[field_name] = spec.fields.pop(field_name)
    return spec


@lru_cache
def _field_name_overrides() -> Dict[str, Dict[str, str]]:
    with open(Path(__file__).resolve().parent / "field-name-overrides.yml", "r") as f:
        return yaml.safe_load(f)


def _postprocess_field_names(spec: Spec) -> Spec:
    spec = deepcopy(spec)
    overrides = _field_name_overrides()
    for field_name, override in overrides.get(spec.name, {}).items():
        spec.fields[field_name].name = override
    return spec


@lru_cache
def _field_type_overrides() -> Dict[str, Dict[str, str]]:
    with open(Path(__file__).resolve().parent / "field-type-overrides.yml", "r") as f:
        return yaml.safe_load(f)


def _postprocess_field_types(spec: Spec) -> Spec:
    spec = deepcopy(spec)

    # Convert type names in the schema to Scitacean class names.
    for field in spec.fields.values():
        for class_name, model_names in _SCHEMA_GROUPS.items():
            for name in model_names:
                if name and name in field.type:
                    field.type = re.sub(name, class_name, field.type)

    overrides = _field_type_overrides()
    for field_name, override in overrides.get(spec.name, {}).items():
        spec.fields[field_name].type = override
    return spec


@lru_cache
def _field_validations() -> Dict[str, Dict[str, str]]:
    with open(Path(__file__).resolve().parent / "field-validations.yml", "r") as f:
        return yaml.safe_load(f)


def _assign_validations(spec: Spec) -> Spec:
    validations = _field_validations()
    if spec.name not in validations:
        return spec

    spec = deepcopy(spec)
    for field_name, validation in validations[spec.name].items():
        if field_name in spec.fields:
            if validation == "size":
                spec.fields[field_name].type = "NonNegativeInt"
            else:
                spec.fields[field_name].validation = validation
    return spec


@lru_cache
def _dataset_field_customizations() -> Dict[str, Any]:
    with open(Path(__file__).resolve().parent / "dataset-fields.yml", "r") as f:
        return yaml.safe_load(f)


def _extend_dataset_fields(spec: DatasetSpec) -> DatasetSpec:
    customizations = _dataset_field_customizations()
    spec = deepcopy(spec)

    for name, field in spec.fields.items():
        field.manual = name in customizations["manual"]
        field.default = customizations["defaults"].get(name)
        if (conv := customizations["conversions"].get(name)) is not None:
            field.conversion = DatasetFieldConversion(**conv)
        if name in customizations["extra_read_only"]:
            field.upload = False
    return spec


def load_specs(schema_url: str) -> Dict[str, Any]:
    schemas = _collect_schemas(load_schemas(schema_url))
    dataset_schema = schemas.pop("Dataset")
    specs = {
        "Dataset": _extend_dataset_fields(
            _build_dataset_spec("Dataset", dataset_schema)
        ),
        **{name: _build_spec(name, updown) for name, updown in schemas.items()},
    }
    return {
        name: _assign_validations(
            _postprocess_field_types(_postprocess_field_names(_mask_fields(spec)))
        )
        for name, spec in specs.items()
    }



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

import pytest

from scitacean import Thumbnail


def test_mime_type():
    thumbnail = Thumbnail(data=b"", mime="image/png")
    assert thumbnail.mime == "image/png"
    assert thumbnail.mime_type == "image"
    assert thumbnail.mime_subtype == "png"


def test_no_mime_type():
    thumbnail = Thumbnail(data=b"", mime=None)
    assert thumbnail.mime is None
    assert thumbnail.mime_type is None
    assert thumbnail.mime_subtype is None


def test_parse_mime_prefix():
    thumbnail = Thumbnail.parse("data:image/png,YWJj")
    assert thumbnail.mime == "image/png"
    assert thumbnail.encoded_data() == "YWJj"


def test_parse_mime_encoding():
    thumbnail = Thumbnail.parse("image/png;base64,YWJj")
    assert thumbnail.mime == "image/png"
    assert thumbnail.encoded_data() == "YWJj"


def test_parse_mime_param():
    thumbnail = Thumbnail.parse("text/html; charset=utf-8,YWEzMzQ=")
    assert thumbnail.mime == "text/html"
    assert thumbnail.encoded_data() == "YWEzMzQ="


def test_parse_mime_2_param2():
    thumbnail = Thumbnail.parse("image/svg+xml; charset=utf-8;base64,eXEzeGE=")
    assert thumbnail.mime == "image/svg+xml"
    assert thumbnail.encoded_data() == "eXEzeGE="


def test_parse_empty_params():
    thumbnail = Thumbnail.parse("image/jpeg;,amY7YQ==")
    assert thumbnail.mime == "image/jpeg"
    assert thumbnail.encoded_data() == "amY7YQ=="


def test_parse_no_params():
    thumbnail = Thumbnail.parse("image/svg,OGZh")
    assert thumbnail.mime == "image/svg"
    assert thumbnail.encoded_data() == "OGZh"


def test_parse_no_mime():
    thumbnail = Thumbnail.parse("amE5MHM4aA==")
    assert thumbnail.mime is None
    assert thumbnail.encoded_data() == "amE5MHM4aA=="


def test_parse_thumbnail_argument():
    thumbnail1 = Thumbnail(mime="image/png", data=b"jak2kcna")
    thumbnail2 = Thumbnail.parse(thumbnail1)
    assert thumbnail1 == thumbnail2
    assert thumbnail1 is not thumbnail2


def test_init_encode():
    thumbnail = Thumbnail(mime=None, data=b"abc")
    assert thumbnail.decoded_data() == b"abc"
    assert thumbnail.encoded_data() == "YWJj"


def test_init_decode():
    thumbnail = Thumbnail(mime=None, _encoded_data="YWEzMzQ=")
    assert thumbnail.decoded_data() == b"aa334"
    assert thumbnail.encoded_data() == "YWEzMzQ="


def test_init_no_data_raises():
    with pytest.raises(TypeError):
        Thumbnail(mime=None)


def test_init_both_data_raises():
    with pytest.raises(TypeError):
        Thumbnail(mime=None, data=b"jalks", _encoded_data="YWEzMzQ=")


def test_load_file(fs):
    fs.create_file("fingers.jpg", contents=b"jal2l9vun2")
    thumbnail = Thumbnail.load_file("fingers.jpg")
    assert thumbnail.mime == "image/jpeg"
    assert thumbnail.decoded_data() == b"jal2l9vun2"
    assert thumbnail.encoded_data() == "amFsMmw5dnVuMg=="


def test_load_file_unknown_mime_type(fs):
    fs.create_file("bad.xxx", contents=b"f9gas03n")
    thumbnail = Thumbnail.load_file("bad.xxx")
    assert thumbnail.mime is None
    assert thumbnail.decoded_data() == b"f9gas03n"
    assert thumbnail.encoded_data() == "ZjlnYXMwM24="


def test_serialize():
    thumbnail = Thumbnail(mime="image/jpeg", data=b"ags9da0")
    assert thumbnail.serialize() == "data:image/jpeg;base64,YWdzOWRhMA=="


def test_serialize_parse_roundtrip():
    thumbnail = Thumbnail(mime="image/svg+xml", data=b"412897a897s")
    serialized = thumbnail.serialize()
    reconstructed = Thumbnail.parse(serialized)
    assert reconstructed == thumbnail



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scitacean contributors (https://github.com/SciCatProject/scitacean)
"""Helpers for running tests with an SFTP server.

This subpackage is primarily meant for testing
:class:`scitacean.testing.sftp.SFTPFileTransfer`.
But it can also be used to test downstream code that uses the SFTP file transfer.

The `pytest <https://docs.pytest.org>`_ fixtures in this package manage an SFTP server
running in a docker container on the local machine.
They, therefore, require docker to be installed and running.

Use the :func:`scitacean.testing.sftp.fixtures.sftp_fileserver`
fixture to manage the server and use
:func:`scitacean.testing.sftp.fixtures.sftp_access`
to get all required access parameters.
See below for examples.

Attention
---------
The fixtures support `pytest-xdist <https://pytest-xdist.readthedocs.io/en/latest/>`_
but only if all workers run on the local machine (the default).

It may still happen that tests fail due to the complexity of synchronizing start up
and shut down of the SFTP server between workers.

See Also
--------
`Testing <../../user-guide/testing.ipynb>`_ user guide.

Examples
--------
In order to test the SFTP file transfer directly, use the provided fixtures to
open a connection manually.
Here, requesting the ``require_sftp_fileserver`` fixture ensures that the server
is running during the test, or that the test gets skipped if SFTP tests are disabled.
Passing the ``connect`` argument as shown ensures that the file transfer
connects to the test server with the correct parameters.

.. code-block:: python

    from scitacean.transfer.sftp import SFTPFileTransfer

    def test_sftp_upload(
        sftp_access,
        sftp_connect_with_username_password,
        require_sftp_fileserver,
        sftp_data_dir,
    ):
        sftp = SFTPFileTransfer(host=sftp_access.host, port=sftp_access.port)
        ds = Dataset(...)
        with sftp.connect_for_upload(
            dataset=ds,
            connect=sftp_connect_with_username_password
        ) as connection:
            # do upload
        # assert that the file has been copied to sftp_data_dir

Testing the SFTP transfer together with a client requires some additional setup.
See ``test_client_with_sftp`` in
`sftp_test.py <https://github.com/SciCatProject/scitacean/blob/main/tests/transfer/sftp_test.py>`_
for an example.

Implementation notes
--------------------
When the server fixture is first used, it initializes the server using these steps:

1. Create a temporary directory with contents::

       tmpdir
         ├ docker-compose.yaml
         ├ .env         (specifies paths for docker volumes)
         ├ counter      (number of workers currently using the server)
         ├ counter.lock (file lock)
         └ data         (storage of files)
            └ seed      (populated from scitacean/testing/sftp/sftp_server_seed)

2. Start docker.
3. Make data writable by the user in docker.
   This changes the ownership of data on the host to root (on some machines).

The docker container and its volumes are removed at the end of the tests.
The fixture also tries to remove the temporary directory.
This can fail as the owner of its contents (in particular data)
may have been changed to root.
So cleanup can fail and leave the directory behind.

Use the seed directory (``sftp_data_dir/"seed"``) to test downloads.
Corresponds to ``/data/seed`` on the server.

Use the base data directory (``sftp_data_dir``) to test uploads.
Corresponds to ``/data`` on the server.

The counter and counter.lock files are used to synchronize starting and stopping
of the docker container between processes.
This is required when ``pytest-xdist`` is used.
Otherwise, those files will not be present.
"""

from ._pytest_helpers import add_pytest_option, sftp_enabled, skip_if_not_sftp
from ._sftp import (
    IgnorePolicy,
    SFTPAccess,
    SFTPUser,
    can_connect,
    configure,
    local_access,
    wait_until_sftp_server_is_live,
)

__all__ = [
    "add_pytest_option",
    "can_connect",
    "configure",
    "local_access",
    "sftp_enabled",
    "skip_if_not_sftp",
    "wait_until_sftp_server_is_live",
    "IgnorePolicy",
    "SFTPAccess",
    "SFTPUser",
]



##########################################
# This file was automatically generated. #
# Do not modify it directly!             #
##########################################

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# flake8: noqa

"""Base class for Dataset."""

from __future__ import annotations

from datetime import datetime, timezone
from typing import Any, Dict, List, Literal, Optional, Tuple, Type, TypeVar, Union

import dateutil.parser

from ._base_model import DatasetType
from ._internal.dataclass_wrapper import dataclass_optional_args
from .datablock import OrigDatablock
from .filesystem import RemotePath
from .model import (
    construct,
    Attachment,
    BaseModel,
    BaseUserModel,
    DownloadDataset,
    DownloadLifecycle,
    Lifecycle,
    Relationship,
    Technique,
)
from .pid import PID


M = TypeVar("M", bound=BaseModel)


def _parse_datetime(x: Optional[Union[datetime, str]]) -> Optional[datetime]:
    if isinstance(x, datetime) or x is None:
        return x
    if x == "now":
        return datetime.now(tz=timezone.utc)
    return dateutil.parser.parse(x)


def _parse_pid(pid: Optional[Union[str, PID]]) -> Optional[PID]:
    if pid is None:
        return pid
    return PID.parse(pid)


def _parse_remote_path(path: Optional[Union[str, RemotePath]]) -> Optional[RemotePath]:
    if path is None:
        return path
    return RemotePath(path)


def _validate_checksum_algorithm(algorithm: Optional[str]) -> Optional[str]:
    if algorithm is None:
        return algorithm
    import hashlib

    if algorithm not in hashlib.algorithms_available:
        raise ValueError(f"Checksum algorithm not recognized: {algorithm}")
    return algorithm


class DatasetBase:
    @dataclass_optional_args(frozen=True, kw_only=True, slots=True)
    class Field:
        name: str
        description: str
        read_only: bool
        required: bool
        scicat_name: str
        type: type
        used_by_derived: bool
        used_by_raw: bool

        def used_by(self, dataset_type: DatasetType) -> bool:
            return (
                self.used_by_raw
                if dataset_type == DatasetType.RAW
                else self.used_by_derived
            )

    _FIELD_SPEC = [
        Field(
            name="type",
            description="Characterize type of dataset, either 'raw' or 'derived'. Autofilled when choosing the proper inherited models.",
            read_only=False,
            required=True,
            scicat_name="type",
            type=DatasetType,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="access_groups",
            description="Optional additional groups which have read access to the data. Users which are members in one of the groups listed here are allowed to access this data. The special group 'public' makes data available to all users.",
            read_only=False,
            required=False,
            scicat_name="accessGroups",
            type=List[str],
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="api_version",
            description="Version of the API used in creation of the dataset.",
            read_only=True,
            required=False,
            scicat_name="version",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="classification",
            description="ACIA information about AUthenticity,COnfidentiality,INtegrity and AVailability requirements of dataset. E.g. AV(ailabilty)=medium could trigger the creation of a two tape copies. Format 'AV=medium,CO=low'",
            read_only=False,
            required=False,
            scicat_name="classification",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="comment",
            description="Comment the user has about a given dataset.",
            read_only=False,
            required=False,
            scicat_name="comment",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="contact_email",
            description="Email of the contact person for this dataset. The string may contain a list of emails, which should then be separated by semicolons.",
            read_only=False,
            required=True,
            scicat_name="contactEmail",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="created_at",
            description="Date and time when this record was created. This property is added and maintained by mongoose.",
            read_only=True,
            required=False,
            scicat_name="createdAt",
            type=datetime,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="created_by",
            description="Indicate the user who created this record. This property is added and maintained by the system.",
            read_only=True,
            required=False,
            scicat_name="createdBy",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="creation_location",
            description="Unique location identifier where data was taken, usually in the form /Site-name/facility-name/instrumentOrBeamline-name. This field is required if the dataset is a Raw dataset.",
            read_only=False,
            required=True,
            scicat_name="creationLocation",
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="creation_time",
            description="Time when dataset became fully available on disk, i.e. all containing files have been written. Format according to chapter 5.6 internet date/time format in RFC 3339. Local times without timezone/offset info are automatically transformed to UTC using the timezone of the API server.",
            read_only=False,
            required=True,
            scicat_name="creationTime",
            type=datetime,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="data_format",
            description="Defines the format of the data files in this dataset, e.g Nexus Version x.y.",
            read_only=False,
            required=False,
            scicat_name="dataFormat",
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="data_quality_metrics",
            description="Data Quality Metrics given by the user to rate the dataset.",
            read_only=False,
            required=False,
            scicat_name="dataQualityMetrics",
            type=int,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="description",
            description="Free text explanation of contents of dataset.",
            read_only=False,
            required=False,
            scicat_name="description",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="end_time",
            description="End time of data acquisition for this dataset, format according to chapter 5.6 internet date/time format in RFC 3339. Local times without timezone/offset info are automatically transformed to UTC using the timezone of the API server.",
            read_only=False,
            required=False,
            scicat_name="endTime",
            type=datetime,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="history",
            description="List of objects containing old and new values.",
            read_only=True,
            required=False,
            scicat_name="history",
            type=type(None),
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="input_datasets",
            description="Array of input dataset identifiers used in producing the derived dataset. Ideally these are the global identifier to existing datasets inside this or federated data catalogs. This field is required if the dataset is a Derived dataset.",
            read_only=False,
            required=True,
            scicat_name="inputDatasets",
            type=List[PID],
            used_by_derived=True,
            used_by_raw=False,
        ),
        Field(
            name="instrument_group",
            description="Optional additional groups which have read and write access to the data. Users which are members in one of the groups listed here are allowed to access this data.",
            read_only=False,
            required=False,
            scicat_name="instrumentGroup",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="instrument_id",
            description="ID of the instrument where the data was created.",
            read_only=False,
            required=False,
            scicat_name="instrumentId",
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="investigator",
            description="First name and last name of the person or people pursuing the data analysis. The string may contain a list of names, which should then be separated by semicolons.",
            read_only=False,
            required=True,
            scicat_name="investigator",
            type=str,
            used_by_derived=True,
            used_by_raw=False,
        ),
        Field(
            name="is_published",
            description="Flag is true when data are made publicly available.",
            read_only=False,
            required=False,
            scicat_name="isPublished",
            type=bool,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="job_log_data",
            description="The output job logfile. Keep the size of this log data well below 15 MB.",
            read_only=False,
            required=False,
            scicat_name="jobLogData",
            type=str,
            used_by_derived=True,
            used_by_raw=False,
        ),
        Field(
            name="job_parameters",
            description="The creation process of the derived data will usually depend on input job parameters. The full structure of these input parameters are stored here.",
            read_only=False,
            required=False,
            scicat_name="jobParameters",
            type=Dict[str, Any],
            used_by_derived=True,
            used_by_raw=False,
        ),
        Field(
            name="keywords",
            description="Array of tags associated with the meaning or contents of this dataset. Values should ideally come from defined vocabularies, taxonomies, ontologies or knowledge graphs.",
            read_only=False,
            required=False,
            scicat_name="keywords",
            type=List[str],
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="license",
            description="Name of the license under which the data can be used.",
            read_only=False,
            required=False,
            scicat_name="license",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="lifecycle",
            description="Describes the current status of the dataset during its lifetime with respect to the storage handling systems.",
            read_only=True,
            required=False,
            scicat_name="datasetlifecycle",
            type=Lifecycle,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="name",
            description="A name for the dataset, given by the creator to carry some semantic meaning. Useful for display purposes e.g. instead of displaying the pid. Will be autofilled if missing using info from sourceFolder.",
            read_only=False,
            required=False,
            scicat_name="datasetName",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="orcid_of_owner",
            description="ORCID of the owner or custodian. The string may contain a list of ORCIDs, which should then be separated by semicolons.",
            read_only=False,
            required=False,
            scicat_name="orcidOfOwner",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="owner",
            description="Owner or custodian of the dataset, usually first name + last name. The string may contain a list of persons, which should then be separated by semicolons.",
            read_only=False,
            required=True,
            scicat_name="owner",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="owner_email",
            description="Email of the owner or custodian of the dataset. The string may contain a list of emails, which should then be separated by semicolons.",
            read_only=False,
            required=False,
            scicat_name="ownerEmail",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="owner_group",
            description="Defines the group which owns the data, and therefore has unrestricted access to this data. Usually a pgroup like p12151",
            read_only=False,
            required=True,
            scicat_name="ownerGroup",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="pid",
            description="Persistent Identifier for datasets derived from UUIDv4 and prepended automatically by site specific PID prefix like 20.500.12345/",
            read_only=True,
            required=False,
            scicat_name="pid",
            type=PID,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="principal_investigator",
            description="First name and last name of principal investigator(s). If multiple PIs are present, use a semicolon separated list. This field is required if the dataset is a Raw dataset.",
            read_only=False,
            required=True,
            scicat_name="principalInvestigator",
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="proposal_id",
            description="The ID of the proposal to which the dataset belongs.",
            read_only=False,
            required=False,
            scicat_name="proposalId",
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="relationships",
            description="Stores the relationships with other datasets.",
            read_only=False,
            required=False,
            scicat_name="relationships",
            type=List[Relationship],
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="sample_id",
            description="ID of the sample used when collecting the data.",
            read_only=False,
            required=False,
            scicat_name="sampleId",
            type=str,
            used_by_derived=False,
            used_by_raw=True,
        ),
        Field(
            name="shared_with",
            description="List of users that the dataset has been shared with.",
            read_only=False,
            required=False,
            scicat_name="sharedWith",
            type=List[str],
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="source_folder",
            description="Absolute file path on file server containing the files of this dataset, e.g. /some/path/to/sourcefolder. In case of a single file dataset, e.g. HDF5 data, it contains the path up to, but excluding the filename. Trailing slashes are removed.",
            read_only=False,
            required=True,
            scicat_name="sourceFolder",
            type=RemotePath,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="source_folder_host",
            description="DNS host name of file server hosting sourceFolder, optionally including a protocol e.g. [protocol://]fileserver1.example.com",
            read_only=False,
            required=False,
            scicat_name="sourceFolderHost",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="techniques",
            description="Stores the metadata information for techniques.",
            read_only=False,
            required=False,
            scicat_name="techniques",
            type=List[Technique],
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="updated_at",
            description="Date and time when this record was updated last. This property is added and maintained by mongoose.",
            read_only=True,
            required=False,
            scicat_name="updatedAt",
            type=datetime,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="updated_by",
            description="Indicate the user who updated this record last. This property is added and maintained by the system.",
            read_only=True,
            required=False,
            scicat_name="updatedBy",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
        Field(
            name="used_software",
            description="A list of links to software repositories which uniquely identifies the pieces of software, including versions, used for yielding the derived data. This field is required if the dataset is a Derived dataset.",
            read_only=False,
            required=True,
            scicat_name="usedSoftware",
            type=List[str],
            used_by_derived=True,
            used_by_raw=False,
        ),
        Field(
            name="validation_status",
            description="Defines a level of trust, e.g. a measure of how much data was verified or used by other persons.",
            read_only=False,
            required=False,
            scicat_name="validationStatus",
            type=str,
            used_by_derived=True,
            used_by_raw=True,
        ),
    ]

    __slots__ = (
        "_access_groups",
        "_api_version",
        "_classification",
        "_comment",
        "_contact_email",
        "_created_at",
        "_created_by",
        "_creation_location",
        "_creation_time",
        "_data_format",
        "_data_quality_metrics",
        "_description",
        "_end_time",
        "_history",
        "_input_datasets",
        "_instrument_group",
        "_instrument_id",
        "_investigator",
        "_is_published",
        "_job_log_data",
        "_job_parameters",
        "_keywords",
        "_license",
        "_lifecycle",
        "_name",
        "_orcid_of_owner",
        "_owner",
        "_owner_email",
        "_owner_group",
        "_pid",
        "_principal_investigator",
        "_proposal_id",
        "_relationships",
        "_sample_id",
        "_shared_with",
        "_source_folder",
        "_source_folder_host",
        "_techniques",
        "_updated_at",
        "_updated_by",
        "_used_software",
        "_validation_status",
        "_meta",
        "_type",
        "_default_checksum_algorithm",
        "_orig_datablocks",
        "_attachments",
    )

    def __init__(
        self,
        type: Union[DatasetType, Literal["raw", "derived"]],
        access_groups: Optional[List[str]] = None,
        classification: Optional[str] = None,
        comment: Optional[str] = None,
        contact_email: Optional[str] = None,
        creation_location: Optional[str] = None,
        creation_time: Optional[Union[str, datetime]] = "now",
        data_format: Optional[str] = None,
        data_quality_metrics: Optional[int] = None,
        description: Optional[str] = None,
        end_time: Optional[datetime] = None,
        input_datasets: Optional[List[PID]] = None,
        instrument_group: Optional[str] = None,
        instrument_id: Optional[str] = None,
        investigator: Optional[str] = None,
        is_published: Optional[bool] = None,
        job_log_data: Optional[str] = None,
        job_parameters: Optional[Dict[str, Any]] = None,
        keywords: Optional[List[str]] = None,
        license: Optional[str] = None,
        name: Optional[str] = None,
        orcid_of_owner: Optional[str] = None,
        owner: Optional[str] = None,
        owner_email: Optional[str] = None,
        owner_group: Optional[str] = None,
        principal_investigator: Optional[str] = None,
        proposal_id: Optional[str] = None,
        relationships: Optional[List[Relationship]] = None,
        sample_id: Optional[str] = None,
        shared_with: Optional[List[str]] = None,
        source_folder: Optional[Union[RemotePath, str]] = None,
        source_folder_host: Optional[str] = None,
        techniques: Optional[List[Technique]] = None,
        used_software: Optional[List[str]] = None,
        validation_status: Optional[str] = None,
        meta: Optional[Dict[str, Any]] = None,
        checksum_algorithm: Optional[str] = "blake2b",
    ) -> None:
        self._type = DatasetType(type)
        self._access_groups = access_groups
        self._classification = classification
        self._comment = comment
        self._contact_email = contact_email
        self._creation_location = creation_location
        self._creation_time = _parse_datetime(creation_time)
        self._data_format = data_format
        self._data_quality_metrics = data_quality_metrics
        self._description = description
        self._end_time = end_time
        self._input_datasets = input_datasets
        self._instrument_group = instrument_group
        self._instrument_id = instrument_id
        self._investigator = investigator
        self._is_published = is_published
        self._job_log_data = job_log_data
        self._job_parameters = job_parameters
        self._keywords = keywords
        self._license = license
        self._name = name
        self._orcid_of_owner = orcid_of_owner
        self._owner = owner
        self._owner_email = owner_email
        self._owner_group = owner_group
        self._principal_investigator = principal_investigator
        self._proposal_id = proposal_id
        self._relationships = relationships
        self._sample_id = sample_id
        self._shared_with = shared_with
        self._source_folder = _parse_remote_path(source_folder)
        self._source_folder_host = source_folder_host
        self._techniques = techniques
        self._used_software = used_software
        self._validation_status = validation_status
        self._api_version = None
        self._created_at = None
        self._created_by = None
        self._history = None
        self._lifecycle = None
        self._pid = None
        self._updated_at = None
        self._updated_by = None
        self._meta = meta or {}
        self._default_checksum_algorithm = _validate_checksum_algorithm(
            checksum_algorithm
        )
        self._orig_datablocks: List[OrigDatablock] = []
        self._attachments: Optional[List[Attachment]] = []

    @property
    def access_groups(self) -> Optional[List[str]]:
        """Optional additional groups which have read access to the data. Users which are members in one of the groups listed here are allowed to access this data. The special group 'public' makes data available to all users."""
        return self._access_groups

    @access_groups.setter
    def access_groups(self, access_groups: Optional[List[str]]) -> None:
        """Optional additional groups which have read access to the data. Users which are members in one of the groups listed here are allowed to access this data. The special group 'public' makes data available to all users."""
        self._access_groups = access_groups

    @property
    def api_version(self) -> Optional[str]:
        """Version of the API used in creation of the dataset."""
        return self._api_version

    @property
    def classification(self) -> Optional[str]:
        """ACIA information about AUthenticity,COnfidentiality,INtegrity and AVailability requirements of dataset. E.g. AV(ailabilty)=medium could trigger the creation of a two tape copies. Format 'AV=medium,CO=low'"""
        return self._classification

    @classification.setter
    def classification(self, classification: Optional[str]) -> None:
        """ACIA information about AUthenticity,COnfidentiality,INtegrity and AVailability requirements of dataset. E.g. AV(ailabilty)=medium could trigger the creation of a two tape copies. Format 'AV=medium,CO=low'"""
        self._classification = classification

    @property
    def comment(self) -> Optional[str]:
        """Comment the user has about a given dataset."""
        return self._comment

    @comment.setter
    def comment(self, comment: Optional[str]) -> None:
        """Comment the user has about a given dataset."""
        self._comment = comment

    @property
    def contact_email(self) -> Optional[str]:
        """Email of the contact person for this dataset. The string may contain a list of emails, which should then be separated by semicolons."""
        return self._contact_email

    @contact_email.setter
    def contact_email(self, contact_email: Optional[str]) -> None:
        """Email of the contact person for this dataset. The string may contain a list of emails, which should then be separated by semicolons."""
        self._contact_email = contact_email

    @property
    def created_at(self) -> Optional[datetime]:
        """Date and time when this record was created. This property is added and maintained by mongoose."""
        return self._created_at

    @property
    def created_by(self) -> Optional[str]:
        """Indicate the user who created this record. This property is added and maintained by the system."""
        return self._created_by

    @property
    def creation_location(self) -> Optional[str]:
        """Unique location identifier where data was taken, usually in the form /Site-name/facility-name/instrumentOrBeamline-name. This field is required if the dataset is a Raw dataset."""
        return self._creation_location

    @creation_location.setter
    def creation_location(self, creation_location: Optional[str]) -> None:
        """Unique location identifier where data was taken, usually in the form /Site-name/facility-name/instrumentOrBeamline-name. This field is required if the dataset is a Raw dataset."""
        self._creation_location = creation_location

    @property
    def creation_time(self) -> Optional[datetime]:
        """Time when dataset became fully available on disk, i.e. all containing files have been written. Format according to chapter 5.6 internet date/time format in RFC 3339. Local times without timezone/offset info are automatically transformed to UTC using the timezone of the API server."""
        return self._creation_time

    @creation_time.setter
    def creation_time(self, creation_time: Optional[Union[str, datetime]]) -> None:
        """Time when dataset became fully available on disk, i.e. all containing files have been written. Format according to chapter 5.6 internet date/time format in RFC 3339. Local times without timezone/offset info are automatically transformed to UTC using the timezone of the API server."""
        self._creation_time = _parse_datetime(creation_time)

    @property
    def data_format(self) -> Optional[str]:
        """Defines the format of the data files in this dataset, e.g Nexus Version x.y."""
        return self._data_format

    @data_format.setter
    def data_format(self, data_format: Optional[str]) -> None:
        """Defines the format of the data files in this dataset, e.g Nexus Version x.y."""
        self._data_format = data_format

    @property
    def data_quality_metrics(self) -> Optional[int]:
        """Data Quality Metrics given by the user to rate the dataset."""
        return self._data_quality_metrics

    @data_quality_metrics.setter
    def data_quality_metrics(self, data_quality_metrics: Optional[int]) -> None:
        """Data Quality Metrics given by the user to rate the dataset."""
        self._data_quality_metrics = data_quality_metrics

    @property
    def description(self) -> Optional[str]:
        """Free text explanation of contents of dataset."""
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        """Free text explanation of contents of dataset."""
        self._description = description

    @property
    def end_time(self) -> Optional[datetime]:
        """End time of data acquisition for this dataset, format according to chapter 5.6 internet date/time format in RFC 3339. Local times without timezone/offset info are automatically transformed to UTC using the timezone of the API server."""
        return self._end_time

    @end_time.setter
    def end_time(self, end_time: Optional[datetime]) -> None:
        """End time of data acquisition for this dataset, format according to chapter 5.6 internet date/time format in RFC 3339. Local times without timezone/offset info are automatically transformed to UTC using the timezone of the API server."""
        self._end_time = end_time

    @property
    def history(self) -> Optional[None]:
        """List of objects containing old and new values."""
        return self._history

    @property
    def input_datasets(self) -> Optional[List[PID]]:
        """Array of input dataset identifiers used in producing the derived dataset. Ideally these are the global identifier to existing datasets inside this or federated data catalogs. This field is required if the dataset is a Derived dataset."""
        return self._input_datasets

    @input_datasets.setter
    def input_datasets(self, input_datasets: Optional[List[PID]]) -> None:
        """Array of input dataset identifiers used in producing the derived dataset. Ideally these are the global identifier to existing datasets inside this or federated data catalogs. This field is required if the dataset is a Derived dataset."""
        self._input_datasets = input_datasets

    @property
    def instrument_group(self) -> Optional[str]:
        """Optional additional groups which have read and write access to the data. Users which are members in one of the groups listed here are allowed to access this data."""
        return self._instrument_group

    @instrument_group.setter
    def instrument_group(self, instrument_group: Optional[str]) -> None:
        """Optional additional groups which have read and write access to the data. Users which are members in one of the groups listed here are allowed to access this data."""
        self._instrument_group = instrument_group

    @property
    def instrument_id(self) -> Optional[str]:
        """ID of the instrument where the data was created."""
        return self._instrument_id

    @instrument_id.setter
    def instrument_id(self, instrument_id: Optional[str]) -> None:
        """ID of the instrument where the data was created."""
        self._instrument_id = instrument_id

    @property
    def investigator(self) -> Optional[str]:
        """First name and last name of the person or people pursuing the data analysis. The string may contain a list of names, which should then be separated by semicolons."""
        return self._investigator

    @investigator.setter
    def investigator(self, investigator: Optional[str]) -> None:
        """First name and last name of the person or people pursuing the data analysis. The string may contain a list of names, which should then be separated by semicolons."""
        self._investigator = investigator

    @property
    def is_published(self) -> Optional[bool]:
        """Flag is true when data are made publicly available."""
        return self._is_published

    @is_published.setter
    def is_published(self, is_published: Optional[bool]) -> None:
        """Flag is true when data are made publicly available."""
        self._is_published = is_published

    @property
    def job_log_data(self) -> Optional[str]:
        """The output job logfile. Keep the size of this log data well below 15 MB."""
        return self._job_log_data

    @job_log_data.setter
    def job_log_data(self, job_log_data: Optional[str]) -> None:
        """The output job logfile. Keep the size of this log data well below 15 MB."""
        self._job_log_data = job_log_data

    @property
    def job_parameters(self) -> Optional[Dict[str, Any]]:
        """The creation process of the derived data will usually depend on input job parameters. The full structure of these input parameters are stored here."""
        return self._job_parameters

    @job_parameters.setter
    def job_parameters(self, job_parameters: Optional[Dict[str, Any]]) -> None:
        """The creation process of the derived data will usually depend on input job parameters. The full structure of these input parameters are stored here."""
        self._job_parameters = job_parameters

    @property
    def keywords(self) -> Optional[List[str]]:
        """Array of tags associated with the meaning or contents of this dataset. Values should ideally come from defined vocabularies, taxonomies, ontologies or knowledge graphs."""
        return self._keywords

    @keywords.setter
    def keywords(self, keywords: Optional[List[str]]) -> None:
        """Array of tags associated with the meaning or contents of this dataset. Values should ideally come from defined vocabularies, taxonomies, ontologies or knowledge graphs."""
        self._keywords = keywords

    @property
    def license(self) -> Optional[str]:
        """Name of the license under which the data can be used."""
        return self._license

    @license.setter
    def license(self, license: Optional[str]) -> None:
        """Name of the license under which the data can be used."""
        self._license = license

    @property
    def lifecycle(self) -> Optional[Lifecycle]:
        """Describes the current status of the dataset during its lifetime with respect to the storage handling systems."""
        return self._lifecycle

    @property
    def name(self) -> Optional[str]:
        """A name for the dataset, given by the creator to carry some semantic meaning. Useful for display purposes e.g. instead of displaying the pid. Will be autofilled if missing using info from sourceFolder."""
        return self._name

    @name.setter
    def name(self, name: Optional[str]) -> None:
        """A name for the dataset, given by the creator to carry some semantic meaning. Useful for display purposes e.g. instead of displaying the pid. Will be autofilled if missing using info from sourceFolder."""
        self._name = name

    @property
    def orcid_of_owner(self) -> Optional[str]:
        """ORCID of the owner or custodian. The string may contain a list of ORCIDs, which should then be separated by semicolons."""
        return self._orcid_of_owner

    @orcid_of_owner.setter
    def orcid_of_owner(self, orcid_of_owner: Optional[str]) -> None:
        """ORCID of the owner or custodian. The string may contain a list of ORCIDs, which should then be separated by semicolons."""
        self._orcid_of_owner = orcid_of_owner

    @property
    def owner(self) -> Optional[str]:
        """Owner or custodian of the dataset, usually first name + last name. The string may contain a list of persons, which should then be separated by semicolons."""
        return self._owner

    @owner.setter
    def owner(self, owner: Optional[str]) -> None:
        """Owner or custodian of the dataset, usually first name + last name. The string may contain a list of persons, which should then be separated by semicolons."""
        self._owner = owner

    @property
    def owner_email(self) -> Optional[str]:
        """Email of the owner or custodian of the dataset. The string may contain a list of emails, which should then be separated by semicolons."""
        return self._owner_email

    @owner_email.setter
    def owner_email(self, owner_email: Optional[str]) -> None:
        """Email of the owner or custodian of the dataset. The string may contain a list of emails, which should then be separated by semicolons."""
        self._owner_email = owner_email

    @property
    def owner_group(self) -> Optional[str]:
        """Defines the group which owns the data, and therefore has unrestricted access to this data. Usually a pgroup like p12151"""
        return self._owner_group

    @owner_group.setter
    def owner_group(self, owner_group: Optional[str]) -> None:
        """Defines the group which owns the data, and therefore has unrestricted access to this data. Usually a pgroup like p12151"""
        self._owner_group = owner_group

    @property
    def pid(self) -> Optional[PID]:
        """Persistent Identifier for datasets derived from UUIDv4 and prepended automatically by site specific PID prefix like 20.500.12345/"""
        return self._pid

    @property
    def principal_investigator(self) -> Optional[str]:
        """First name and last name of principal investigator(s). If multiple PIs are present, use a semicolon separated list. This field is required if the dataset is a Raw dataset."""
        return self._principal_investigator

    @principal_investigator.setter
    def principal_investigator(self, principal_investigator: Optional[str]) -> None:
        """First name and last name of principal investigator(s). If multiple PIs are present, use a semicolon separated list. This field is required if the dataset is a Raw dataset."""
        self._principal_investigator = principal_investigator

    @property
    def proposal_id(self) -> Optional[str]:
        """The ID of the proposal to which the dataset belongs."""
        return self._proposal_id

    @proposal_id.setter
    def proposal_id(self, proposal_id: Optional[str]) -> None:
        """The ID of the proposal to which the dataset belongs."""
        self._proposal_id = proposal_id

    @property
    def relationships(self) -> Optional[List[Relationship]]:
        """Stores the relationships with other datasets."""
        return self._relationships

    @relationships.setter
    def relationships(self, relationships: Optional[List[Relationship]]) -> None:
        """Stores the relationships with other datasets."""
        self._relationships = relationships

    @property
    def sample_id(self) -> Optional[str]:
        """ID of the sample used when collecting the data."""
        return self._sample_id

    @sample_id.setter
    def sample_id(self, sample_id: Optional[str]) -> None:
        """ID of the sample used when collecting the data."""
        self._sample_id = sample_id

    @property
    def shared_with(self) -> Optional[List[str]]:
        """List of users that the dataset has been shared with."""
        return self._shared_with

    @shared_with.setter
    def shared_with(self, shared_with: Optional[List[str]]) -> None:
        """List of users that the dataset has been shared with."""
        self._shared_with = shared_with

    @property
    def source_folder(self) -> Optional[RemotePath]:
        """Absolute file path on file server containing the files of this dataset, e.g. /some/path/to/sourcefolder. In case of a single file dataset, e.g. HDF5 data, it contains the path up to, but excluding the filename. Trailing slashes are removed."""
        return self._source_folder

    @source_folder.setter
    def source_folder(self, source_folder: Optional[Union[RemotePath, str]]) -> None:
        """Absolute file path on file server containing the files of this dataset, e.g. /some/path/to/sourcefolder. In case of a single file dataset, e.g. HDF5 data, it contains the path up to, but excluding the filename. Trailing slashes are removed."""
        self._source_folder = _parse_remote_path(source_folder)

    @property
    def source_folder_host(self) -> Optional[str]:
        """DNS host name of file server hosting sourceFolder, optionally including a protocol e.g. [protocol://]fileserver1.example.com"""
        return self._source_folder_host

    @source_folder_host.setter
    def source_folder_host(self, source_folder_host: Optional[str]) -> None:
        """DNS host name of file server hosting sourceFolder, optionally including a protocol e.g. [protocol://]fileserver1.example.com"""
        self._source_folder_host = source_folder_host

    @property
    def techniques(self) -> Optional[List[Technique]]:
        """Stores the metadata information for techniques."""
        return self._techniques

    @techniques.setter
    def techniques(self, techniques: Optional[List[Technique]]) -> None:
        """Stores the metadata information for techniques."""
        self._techniques = techniques

    @property
    def updated_at(self) -> Optional[datetime]:
        """Date and time when this record was updated last. This property is added and maintained by mongoose."""
        return self._updated_at

    @property
    def updated_by(self) -> Optional[str]:
        """Indicate the user who updated this record last. This property is added and maintained by the system."""
        return self._updated_by

    @property
    def used_software(self) -> Optional[List[str]]:
        """A list of links to software repositories which uniquely identifies the pieces of software, including versions, used for yielding the derived data. This field is required if the dataset is a Derived dataset."""
        return self._used_software

    @used_software.setter
    def used_software(self, used_software: Optional[List[str]]) -> None:
        """A list of links to software repositories which uniquely identifies the pieces of software, including versions, used for yielding the derived data. This field is required if the dataset is a Derived dataset."""
        self._used_software = used_software

    @property
    def validation_status(self) -> Optional[str]:
        """Defines a level of trust, e.g. a measure of how much data was verified or used by other persons."""
        return self._validation_status

    @validation_status.setter
    def validation_status(self, validation_status: Optional[str]) -> None:
        """Defines a level of trust, e.g. a measure of how much data was verified or used by other persons."""
        self._validation_status = validation_status

    @property
    def meta(self) -> Dict[str, Any]:
        """Dict of scientific metadata."""
        return self._meta

    @meta.setter
    def meta(self, meta: Dict[str, Any]) -> None:
        """Dict of scientific metadata."""
        self._meta = meta

    @property
    def type(self) -> DatasetType:
        """Characterize type of dataset, either 'raw' or 'derived'. Autofilled when choosing the proper inherited models."""
        return self._type

    @staticmethod
    def _prepare_fields_from_download(
        download_model: DownloadDataset,
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        init_args = {}
        read_only = {}
        for field in DatasetBase._FIELD_SPEC:
            if field.read_only:
                read_only["_" + field.name] = getattr(download_model, field.scicat_name)
            else:
                init_args[field.name] = getattr(download_model, field.scicat_name)

        init_args["meta"] = download_model.scientificMetadata
        _convert_download_fields_in_place(init_args, read_only)

        return init_args, read_only

    @staticmethod
    def _convert_readonly_fields_in_place(read_only: Dict[str, Any]) -> None:
        if (pid := read_only.get("_pid")) is not None:
            read_only["_pid"] = _parse_pid(pid)


def _convert_download_fields_in_place(
    init_args: Dict[str, Any], read_only: Dict[str, Any]
) -> None:
    for mod, key in ((Technique, "techniques"), (Relationship, "relationships")):
        init_args[key] = _list_field_from_download(mod, init_args.get(key))

    DatasetBase._convert_readonly_fields_in_place(read_only)
    if (lifecycle := read_only.get("_lifecycle")) is not None:
        read_only["_lifecycle"] = Lifecycle.from_download_model(
            _as_model(DownloadLifecycle, lifecycle)
        )


def _list_field_from_download(
    mod: Type[BaseUserModel], value: Optional[List[Any]]
) -> Optional[List[BaseUserModel]]:
    if value is None:
        return None
    return [
        mod.from_download_model(_as_model(mod.download_model_type(), item))
        for item in value
    ]


# If validation fails, sub models are not converted automatically by Pydantic.
def _as_model(mod: Type[M], value: Union[M, Dict[str, Any]]) -> M:
    if isinstance(value, dict):
        return construct(mod, **value, _strict_validation=False)
    return value



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

from pathlib import Path
from typing import Optional, Tuple, Union

import pytest

from .._internal.file_counter import FileCounter, NullCounter


def using_xdist(request: pytest.FixtureRequest) -> bool:
    """Return True if we are running with one or more pytest-xdist workers."""
    try:
        is_master = request.getfixturevalue("worker_id") == "master"
        return not is_master
    except pytest.FixtureLookupError:
        return False


def root_tmp_dir(
    request: pytest.FixtureRequest, tmp_path_factory: pytest.TempPathFactory
) -> Path:
    """Get the temp directory for this invocation of pytest shared by all workers."""
    if using_xdist(request):
        return tmp_path_factory.getbasetemp().parent
    return tmp_path_factory.getbasetemp()


def init_pytest_work_dir(
    request: pytest.FixtureRequest,
    tmp_path_factory: pytest.TempPathFactory,
    name: Optional[str],
) -> Tuple[Path, Union[FileCounter, NullCounter]]:
    """Create a working directory and initialize an atomic counter and lock for it."""
    return init_work_dir(request, root_tmp_dir(request, tmp_path_factory), name)


def init_work_dir(
    request: pytest.FixtureRequest, base_path: Path, name: Optional[str]
) -> Tuple[Path, Union[FileCounter, NullCounter]]:
    """Create a working directory and initialize an atomic counter and lock for it."""
    target_dir = base_path / name if name else base_path
    target_dir.mkdir(exist_ok=True)

    counter: Union[FileCounter, NullCounter]
    if using_xdist(request):
        counter = FileCounter(target_dir / "counter")
    else:
        counter = NullCounter()

    return target_dir, counter



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
from typing import Any, Callable, Literal

import pydantic


def is_pydantic_v1() -> bool:
    if pydantic.__version__.split(".", 1)[0] == "1":
        import warnings

        from ..warning import VisibleDeprecationWarning

        warnings.warn(
            "Support for Pydantic v1 is deprecated in Scitacean"
            "23.10.0 and will be removed in Scitacean v23.12.0."
            "If you cannot update your Pydantic version, please comment in this issue:"
            " https://github.com/SciCatProject/scitacean/issues/158",
            VisibleDeprecationWarning,
            stacklevel=2,
        )
        return True

    return False


def field_validator(
    *args: Any,
    mode: Literal["before", "after", "wrap", "plain"] = "after",
    **kwargs: Any,
) -> Callable[[Any], Any]:
    if is_pydantic_v1():
        return pydantic.validator(*args, pre=(mode == "before"), **kwargs)
    return pydantic.field_validator(*args, mode=mode, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)

import pytest

from scitacean import PID, Dataset
from scitacean.util.formatter import DatasetPathFormatter


def test_dataset_formatter_uses_dataset_fields():
    dset = Dataset(type="raw", owner="magrat")
    fmt = "{type} {owner}"
    formatted = DatasetPathFormatter().format(fmt, dset=dset)
    expected = fmt.format(type=dset.type, owner=dset.owner)
    assert formatted == expected


def test_dataset_formatter_can_access_attrs_of_fields():
    dset = Dataset(type="raw")
    dset._pid = PID.parse("prefix/actual-id")  # type: ignore[assignment]
    formatted = DatasetPathFormatter().format("{pid.pid}", dset=dset)
    assert formatted == "actual-id"


def test_dataset_formatter_supports_special_uid():
    dset = Dataset(type="raw", owner="magrat")
    fmt = "id: {uid}"

    formatted = DatasetPathFormatter().format(fmt, uid="super-unique")
    expected = "id: super-unique"
    assert formatted == expected

    formatted = DatasetPathFormatter().format(fmt, dset=dset, uid="super-unique")
    expected = "id: super-unique"
    assert formatted == expected


def test_dataset_formatter_supports_dset_and_uid():
    dset = Dataset(type="raw", owner="ogg")
    fmt = "{uid}({owner})"

    formatted = DatasetPathFormatter().format(fmt, dset=dset, uid="abcd")
    expected = "abcd(ogg)"
    assert formatted == expected


def test_dataset_formatter_escapes_characters():
    dset = Dataset(type="raw", owner="Harald Blåtand", owner_email="harald@viking.dk")
    formatted = DatasetPathFormatter().format("{owner}-{owner_email}", dset=dset)
    expected = "Harald Bl_xe5tand-harald_viking.dk"
    assert formatted == expected


def test_dataset_formatter_preserves_path_separators():
    dset = Dataset(type="raw", owner="Nanny Ogg")
    formatted = DatasetPathFormatter().format("{type}/{owner}.data", dset=dset)
    assert formatted == "raw/Nanny Ogg.data"


def test_dataset_formatter_does_not_allow_none():
    dset = Dataset(type="raw", owner=None)
    with pytest.raises(ValueError):
        DatasetPathFormatter().format("{owner}", dset=dset)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
# mypy: disable-error-code="no-untyped-def"
"""Pytest fixtures to manage and access a local SSH server."""

import logging
from pathlib import Path
from typing import Callable, Generator, Optional

import fabric
import fabric.config
import pytest

from ..._internal import docker
from .._pytest_helpers import init_work_dir, root_tmp_dir
from ._pytest_helpers import skip_if_not_ssh, ssh_enabled
from ._ssh import (
    IgnorePolicy,
    SSHAccess,
    configure,
    local_access,
    wait_until_ssh_server_is_live,
)


@pytest.fixture(scope="session")
def ssh_access(request: pytest.FixtureRequest) -> SSHAccess:
    """Fixture that returns SSH access parameters.

    Returns
    -------
    :
        A URL and user to connect to the testing SSH server.
        The user has access to all initial files registered in the
        database and permissions to create new files.
    """
    skip_if_not_ssh(request)
    return local_access()


@pytest.fixture(scope="session")
def ssh_base_dir(
    request: pytest.FixtureRequest, tmp_path_factory: pytest.TempPathFactory
) -> Optional[Path]:
    """Fixture that returns the base working directory for the SSH server setup.

    Returns
    -------
    :
        A path to a directory on the host machine.
        The directory gets populated by the
        :func:`scitacean.testing.ssh.fixtures.ssh_fileserver` fixture.
        It contains the docker configuration and volumes.

        Returns ``None`` if SSH tests are disabled
    """
    if not ssh_enabled(request):
        return None
    return root_tmp_dir(request, tmp_path_factory) / "scitacean-ssh"


@pytest.fixture(scope="session")
def ssh_data_dir(ssh_base_dir: Optional[Path]) -> Optional[Path]:
    """Fixture that returns the data directory for the SSH server setup.

    Returns
    -------
    :
        A path to a directory on the host machine.
        The directory is mounted as ``/data`` on the server.

        Returns ``None`` if SSH tests are disabled
    """
    if ssh_base_dir is None:
        return None
    return ssh_base_dir / "data"


@pytest.fixture()
def require_ssh_fileserver(request, ssh_fileserver) -> None:
    """Fixture to declare that a test needs a local SSH server.

    Like :func:`scitacean.testing.ssh.ssh_fileserver`
    but this skips the test if SSH tests are disabled.
    """
    skip_if_not_ssh(request)


@pytest.fixture(scope="session")
def ssh_fileserver(
    request: pytest.FixtureRequest,
    ssh_access: SSHAccess,
    ssh_base_dir: Optional[Path],
    ssh_data_dir: Optional[Path],
    ssh_connect_with_username_password,
) -> Generator[bool, None, None]:
    """Fixture to declare that a test needs a local SSH server.

    If SSH tests are enabled, this fixture configures and starts an SSH server
    in a docker container the first time a test requests it.
    The server and container will be stopped and removed at the end of the test session.

    Does nothing if the SSH tests are disabled.

    Returns
    -------
    :
        True if SSH tests are enabled and False otherwise.
    """
    if ssh_base_dir is None:
        yield False
        return

    target_dir, counter = init_work_dir(request, ssh_base_dir, name=None)

    try:
        with counter.increment() as count:
            if count == 1:
                _ssh_docker_up(target_dir, ssh_access)
            elif not _ssh_server_is_running():
                raise RuntimeError("Expected SSH server to be running")
        yield True
    finally:
        with counter.decrement() as count:
            if count == 0:
                _ssh_docker_down(target_dir)


@pytest.fixture(scope="session")
def ssh_connection_config() -> fabric.config.Config:
    """Fixture that returns the configuration for fabric.Connection for tests.

    Can be used to open SSH connections if ``SSHFileTransfer`` is not enough.
    """
    config = fabric.config.Config()
    config["load_ssh_configs"] = False
    config["connect_kwargs"] = {
        "allow_agent": False,
        "look_for_keys": False,
    }
    return config


@pytest.fixture(scope="session")
def ssh_connect_with_username_password(
    ssh_access: SSHAccess, ssh_connection_config: fabric.config.Config
) -> Callable[..., fabric.Connection]:
    """Fixture that returns a function to create a connection to the testing SSH server.

    Uses username+password and rejects any other authentication attempt.

    Returns
    -------
    :
        A function to pass as the ``connect`` argument to
        :meth:`scitacean.transfer.ssh.SSHFileTransfer.connect_for_download`
        or :meth:`scitacean.transfer.ssh.SSHFileTransfer.connect_for_upload`.

    Examples
    --------
    Explicitly connect to the

    .. code-block:: python

        def test_ssh(ssh_access, ssh_connect_with_username_password, ssh_fileserver):
            ssh = SSHFileTransfer(host=ssh_access.host, port=ssh_access.port)
            with ssh.connect_for_download(
                connect=ssh_connect_with_username_password
            ) as connection:
                # use connection
    """

    def connect(host: str, port: int, **kwargs):
        if kwargs:
            raise ValueError(
                "ssh_connect_with_username_password must only be"
                f" used without extra arguments. Got {kwargs=}"
            )
        connection = fabric.Connection(
            host=host,
            port=port,
            user=ssh_access.user.username,
            config=ssh_connection_config,
            connect_kwargs={
                "password": ssh_access.user.password,
                **ssh_connection_config.connect_kwargs,
            },
        )
        connection.client.set_missing_host_key_policy(IgnorePolicy())
        return connection

    return connect


def _ssh_docker_up(target_dir: Path, ssh_access: SSHAccess) -> None:
    if _ssh_server_is_running():
        raise RuntimeError("SSH docker container is already running")
    docker_compose_file = target_dir / "docker-compose.yaml"
    log = logging.getLogger("scitacean.testing")
    log.info("Starting docker container with SSH server from %s", docker_compose_file)
    configure(target_dir)
    docker.docker_compose_up(docker_compose_file)
    log.info("Waiting for SSH docker to become accessible")
    wait_until_ssh_server_is_live(ssh_access=ssh_access, max_time=20, n_tries=20)
    log.info("Successfully connected to SSH server")
    # Give the user write access.
    docker.docker_compose_run(
        docker_compose_file, "scitacean-test-ssh-server", "chown", "1000:1000", "/data"
    )


def _ssh_docker_down(target_dir: Path) -> None:
    # Check if container is running because the fixture can call this function
    # if there was an exception in _ssh_docker_up.
    # In that case, there is nothing to tear down.
    if _ssh_server_is_running():
        docker_compose_file = target_dir / "docker-compose.yaml"
        log = logging.getLogger("scitacean.testing")
        log.info(
            "Stopping docker container with SSH server from %s", docker_compose_file
        )
        docker.docker_compose_down(docker_compose_file)


def _ssh_server_is_running() -> bool:
    return docker.container_is_running("scitacean-test-ssh")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
import importlib.resources
import os
import time
from copy import deepcopy
from typing import Any, Dict, Union
from urllib.parse import urljoin

import requests
import yaml

from ..._internal.docker import docker_compose_down, docker_compose_up
from . import config

_PathLike = Union[str, os.PathLike]


def _read_yaml(filename: str) -> Any:
    if hasattr(importlib.resources, "files"):
        # Use new API added in Python 3.9
        return yaml.safe_load(
            importlib.resources.files("scitacean.testing.backend")
            .joinpath(filename)
            .read_text()
        )
    # Old API, deprecated as of Python 3.11
    return yaml.safe_load(
        importlib.resources.read_text("scitacean.testing.backend", filename)
    )


def _docker_compose_template() -> Dict[str, Any]:
    template = _read_yaml("docker-compose-backend-template.yaml")
    return template  # type: ignore[no-any-return]


def _apply_config(template: Dict[str, Any]) -> Dict[str, Any]:
    res = deepcopy(template)
    scicat = res["services"]["scicat"]
    ports = scicat["ports"][0].split(":")
    scicat["ports"] = [f"{ports[0]}:{config.SCICAT_PORT}"]

    env = scicat["environment"]
    env["PORT"] = config.SCICAT_PORT
    env["PID_PREFIX"] = config.PID_PREFIX
    env["SITE"] = config.SITE

    return res


def configure(target_path: _PathLike) -> None:
    """Build a docker-compose file for the testing backend.

    Parameters
    ----------
    target_path:
        Generate a docker-compose file at this path.
    """
    c = yaml.dump(_apply_config(_docker_compose_template()))
    if "PLACEHOLDER" in c:
        raise RuntimeError("Incorrect config")

    with open(target_path, "w") as f:
        f.write(c)


def start_backend(docker_compose_file: _PathLike) -> None:
    """Start the docker container with SciCat backend."""
    docker_compose_up(docker_compose_file)


def stop_backend(docker_compose_file: _PathLike) -> None:
    """Stop the docker container with SciCat backend and remove volumes."""
    docker_compose_down(docker_compose_file)


def can_connect() -> bool:
    """Test the connection to the testing SciCat backend.

    Returns
    -------
    :
        True if the backend is reachable, False otherwise.
    """
    scicat_access = config.local_access("user1")
    try:
        response = requests.post(
            urljoin(scicat_access.url, "Users/login"),
            json=scicat_access.user.credentials,
            timeout=0.5,
        )
    except requests.ConnectionError:
        return False
    return response.ok


def wait_until_backend_is_live(max_time: float, n_tries: int) -> None:
    """Sleep until a connection to the backend can be made.

    The backend takes a few seconds to become usable after the
    docker container was started.
    This function attempts to connect periodically until the connection
    succeeds or ``max_time`` is reached.

    Parameters
    ----------
    max_time:
        Maximum time in seconds to wait for the backend to become usable.
    n_tries:
        Number of connection attempts within ``max_time``.

    Raises
    ------
    RuntimeError
        If no connection can be made within the time limit.
    """
    for _ in range(n_tries):
        if can_connect():
            return
        time.sleep(max_time / n_tries)
    if not can_connect():
        raise RuntimeError("Cannot connect to backend")



##########################################
# This file was automatically generated. #
# Do not modify it directly!             #
##########################################

# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Models for communication with SciCat and user facing dataclasses.

The high-level :class:`scitacean.Client` and :class:`scitacean.Dataset` return objects
from the SciCat database in the form of 'user models'.
Those are usually all that is required for working with Scitacean.

At a lower level, those models are converted to or from upload or download models,
respectively, by the corresponding methods of the user models.
These upload and download models represent SciCat's schemas more closely and are used
by the lower-level :class:`scitacean.client.ScicatClient`
and :class:`scitacean.testing.client.FakeClient`.

.. rubric:: User models

Dataclasses exposed to users, primarily through :class:`Dataset`.

.. autosummary::
  :toctree: ../classes
  :template: scitacean-class-template.rst

  Attachment
  DatasetType
  History
  Instrument
  Lifecycle
  Relationship
  Sample
  Technique

.. rubric:: Download models

Pydantic models for the data received from SciCat in downloads.

.. autosummary::
  :toctree: ../classes

  DownloadAttachment
  DownloadDatablock
  DownloadDataFile
  DownloadDataset
  DownloadHistory
  DownloadInstrument
  DownloadLifecycle
  DownloadOrigDatablock
  DownloadRelationship
  DownloadSample
  DownloadTechnique

.. rubric:: Upload models

Pydantic models sent to SciCat in uploads.

.. autosummary::
  :toctree: ../classes

  UploadAttachment
  UploadDatablock
  UploadDataFile
  UploadDerivedDataset
  UploadOrigDatablock
  UploadRawDataset
  UploadRelationship
  UploadSample
  UploadTechnique

.. rubric:: Functions

.. autosummary::
   :toctree: ../functions

   construct
"""

from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, List, Optional, Type

import pydantic
from pydantic import NonNegativeInt

from ._base_model import (
    BaseModel,
    BaseUserModel,
    DatasetType,
    construct,
    validate_datetime,
    validate_drop,
    validate_emails,
    validate_orcids,
)
from ._internal.dataclass_wrapper import dataclass_optional_args
from ._internal.pydantic_compat import field_validator
from .filesystem import RemotePath
from .pid import PID
from .thumbnail import Thumbnail


class DownloadDataset(
    BaseModel, masked=("attachments", "datablocks", "origdatablocks")
):
    contactEmail: Optional[str] = None
    creationLocation: Optional[str] = None
    creationTime: Optional[datetime] = None
    inputDatasets: Optional[List[PID]] = None
    investigator: Optional[str] = None
    numberOfFilesArchived: Optional[NonNegativeInt] = None
    owner: Optional[str] = None
    ownerGroup: Optional[str] = None
    principalInvestigator: Optional[str] = None
    sourceFolder: Optional[RemotePath] = None
    type: Optional[DatasetType] = None
    usedSoftware: Optional[List[str]] = None
    accessGroups: Optional[List[str]] = None
    version: Optional[str] = None
    classification: Optional[str] = None
    comment: Optional[str] = None
    createdAt: Optional[datetime] = None
    createdBy: Optional[str] = None
    dataFormat: Optional[str] = None
    dataQualityMetrics: Optional[int] = None
    description: Optional[str] = None
    endTime: Optional[datetime] = None
    history: Optional[None] = None
    instrumentGroup: Optional[str] = None
    instrumentId: Optional[str] = None
    isPublished: Optional[bool] = None
    jobLogData: Optional[str] = None
    jobParameters: Optional[Dict[str, Any]] = None
    keywords: Optional[List[str]] = None
    license: Optional[str] = None
    datasetlifecycle: Optional[DownloadLifecycle] = None
    scientificMetadata: Optional[Dict[str, Any]] = None
    datasetName: Optional[str] = None
    numberOfFiles: Optional[NonNegativeInt] = None
    orcidOfOwner: Optional[str] = None
    ownerEmail: Optional[str] = None
    packedSize: Optional[NonNegativeInt] = None
    pid: Optional[PID] = None
    proposalId: Optional[str] = None
    relationships: Optional[List[DownloadRelationship]] = None
    sampleId: Optional[str] = None
    sharedWith: Optional[List[str]] = None
    size: Optional[NonNegativeInt] = None
    sourceFolderHost: Optional[str] = None
    techniques: Optional[List[DownloadTechnique]] = None
    updatedAt: Optional[datetime] = None
    updatedBy: Optional[str] = None
    validationStatus: Optional[str] = None

    @field_validator("creationTime", "createdAt", "endTime", "updatedAt", mode="before")
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @field_validator("history", mode="before")
    def _validate_drop(cls, value: Any) -> Any:
        return validate_drop(value)

    @field_validator("contactEmail", "ownerEmail", mode="before")
    def _validate_emails(cls, value: Any) -> Any:
        return validate_emails(value)

    @field_validator("orcidOfOwner", mode="before")
    def _validate_orcids(cls, value: Any) -> Any:
        return validate_orcids(value)


class UploadDerivedDataset(BaseModel):
    contactEmail: str
    creationTime: datetime
    inputDatasets: List[PID]
    investigator: str
    numberOfFilesArchived: NonNegativeInt
    owner: str
    ownerGroup: str
    sourceFolder: RemotePath
    type: DatasetType
    usedSoftware: List[str]
    accessGroups: Optional[List[str]] = None
    classification: Optional[str] = None
    comment: Optional[str] = None
    dataQualityMetrics: Optional[int] = None
    description: Optional[str] = None
    instrumentGroup: Optional[str] = None
    isPublished: Optional[bool] = None
    jobLogData: Optional[str] = None
    jobParameters: Optional[Dict[str, Any]] = None
    keywords: Optional[List[str]] = None
    license: Optional[str] = None
    scientificMetadata: Optional[Dict[str, Any]] = None
    datasetName: Optional[str] = None
    numberOfFiles: Optional[NonNegativeInt] = None
    orcidOfOwner: Optional[str] = None
    ownerEmail: Optional[str] = None
    packedSize: Optional[NonNegativeInt] = None
    relationships: Optional[List[UploadRelationship]] = None
    sharedWith: Optional[List[str]] = None
    size: Optional[NonNegativeInt] = None
    sourceFolderHost: Optional[str] = None
    techniques: Optional[List[UploadTechnique]] = None
    validationStatus: Optional[str] = None

    @field_validator("creationTime", mode="before")
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @field_validator("contactEmail", "ownerEmail", mode="before")
    def _validate_emails(cls, value: Any) -> Any:
        return validate_emails(value)

    @field_validator("orcidOfOwner", mode="before")
    def _validate_orcids(cls, value: Any) -> Any:
        return validate_orcids(value)


class UploadRawDataset(BaseModel):
    contactEmail: str
    creationLocation: str
    creationTime: datetime
    numberOfFilesArchived: NonNegativeInt
    owner: str
    ownerGroup: str
    principalInvestigator: str
    sourceFolder: RemotePath
    type: DatasetType
    accessGroups: Optional[List[str]] = None
    classification: Optional[str] = None
    comment: Optional[str] = None
    dataFormat: Optional[str] = None
    dataQualityMetrics: Optional[int] = None
    description: Optional[str] = None
    endTime: Optional[datetime] = None
    instrumentGroup: Optional[str] = None
    instrumentId: Optional[str] = None
    isPublished: Optional[bool] = None
    keywords: Optional[List[str]] = None
    license: Optional[str] = None
    scientificMetadata: Optional[Dict[str, Any]] = None
    datasetName: Optional[str] = None
    numberOfFiles: Optional[NonNegativeInt] = None
    orcidOfOwner: Optional[str] = None
    ownerEmail: Optional[str] = None
    packedSize: Optional[NonNegativeInt] = None
    proposalId: Optional[str] = None
    relationships: Optional[List[UploadRelationship]] = None
    sampleId: Optional[str] = None
    sharedWith: Optional[List[str]] = None
    size: Optional[NonNegativeInt] = None
    sourceFolderHost: Optional[str] = None
    techniques: Optional[List[UploadTechnique]] = None
    validationStatus: Optional[str] = None

    @field_validator("creationTime", "endTime", mode="before")
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @field_validator("contactEmail", "ownerEmail", mode="before")
    def _validate_emails(cls, value: Any) -> Any:
        return validate_emails(value)

    @field_validator("orcidOfOwner", mode="before")
    def _validate_orcids(cls, value: Any) -> Any:
        return validate_orcids(value)


class DownloadAttachment(BaseModel):
    caption: Optional[str] = None
    ownerGroup: Optional[str] = None
    accessGroups: Optional[List[str]] = None
    createdAt: Optional[datetime] = None
    createdBy: Optional[str] = None
    datasetId: Optional[PID] = None
    id: Optional[str] = None
    instrumentGroup: Optional[str] = None
    proposalId: Optional[str] = None
    sampleId: Optional[str] = None
    thumbnail: Optional[Thumbnail] = None
    updatedAt: Optional[datetime] = None
    updatedBy: Optional[str] = None

    @field_validator("createdAt", "updatedAt", mode="before")
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @classmethod
    def user_model_type(cls) -> Type[Attachment]:
        return Attachment

    @classmethod
    def upload_model_type(cls) -> Type[UploadAttachment]:
        return UploadAttachment


class UploadAttachment(BaseModel):
    caption: str
    ownerGroup: str
    accessGroups: Optional[List[str]] = None
    datasetId: Optional[PID] = None
    id: Optional[str] = None
    instrumentGroup: Optional[str] = None
    proposalId: Optional[str] = None
    sampleId: Optional[str] = None
    thumbnail: Optional[Thumbnail] = None

    @classmethod
    def user_model_type(cls) -> Type[Attachment]:
        return Attachment

    @classmethod
    def download_model_type(cls) -> Type[DownloadAttachment]:
        return DownloadAttachment


class DownloadOrigDatablock(BaseModel):
    dataFileList: Optional[List[DownloadDataFile]] = None
    datasetId: Optional[PID] = None
    size: Optional[NonNegativeInt] = None
    id: Optional[str] = pydantic.Field(alias="_id", default=None)
    accessGroups: Optional[List[str]] = None
    chkAlg: Optional[str] = None
    createdAt: Optional[datetime] = None
    createdBy: Optional[str] = None
    instrumentGroup: Optional[str] = None
    ownerGroup: Optional[str] = None
    updatedAt: Optional[datetime] = None
    updatedBy: Optional[str] = None

    @field_validator("createdAt", "updatedAt", mode="before")
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @classmethod
    def upload_model_type(cls) -> Type[UploadOrigDatablock]:
        return UploadOrigDatablock


class UploadOrigDatablock(BaseModel):
    dataFileList: List[UploadDataFile]
    datasetId: PID
    size: NonNegativeInt
    accessGroups: Optional[List[str]] = None
    chkAlg: Optional[str] = None
    instrumentGroup: Optional[str] = None
    ownerGroup: Optional[str] = None

    @classmethod
    def download_model_type(cls) -> Type[DownloadOrigDatablock]:
        return DownloadOrigDatablock


class DownloadDatablock(BaseModel):
    archiveId: Optional[str] = None
    dataFileList: Optional[List[DownloadDataFile]] = None
    packedSize: Optional[NonNegativeInt] = None
    size: Optional[NonNegativeInt] = None
    version: Optional[str] = None
    id: Optional[str] = pydantic.Field(alias="_id", default=None)
    accessGroups: Optional[List[str]] = None
    chkAlg: Optional[str] = None
    createdAt: Optional[datetime] = None
    createdBy: Optional[str] = None
    datasetId: Optional[PID] = None
    instrumentGroup: Optional[str] = None
    ownerGroup: Optional[str] = None
    updatedAt: Optional[datetime] = None
    updatedBy: Optional[str] = None

    @field_validator("createdAt", "updatedAt", mode="before")
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @classmethod
    def upload_model_type(cls) -> Type[UploadDatablock]:
        return UploadDatablock


class UploadDatablock(BaseModel):
    archiveId: str
    dataFileList: List[UploadDataFile]
    packedSize: NonNegativeInt
    size: NonNegativeInt
    version: str
    chkAlg: Optional[str] = None

    @classmethod
    def download_model_type(cls) -> Type[DownloadDatablock]:
        return DownloadDatablock


class DownloadLifecycle(BaseModel):
    archivable: Optional[bool] = None
    archiveRetentionTime: Optional[datetime] = None
    archiveReturnMessage: Optional[Dict[str, Any]] = None
    archiveStatusMessage: Optional[str] = None
    dateOfDiskPurging: Optional[datetime] = None
    dateOfPublishing: Optional[datetime] = None
    exportedTo: Optional[str] = None
    isOnCentralDisk: Optional[bool] = None
    publishable: Optional[bool] = None
    publishedOn: Optional[datetime] = None
    retrievable: Optional[bool] = None
    retrieveIntegrityCheck: Optional[bool] = None
    retrieveReturnMessage: Optional[Dict[str, Any]] = None
    retrieveStatusMessage: Optional[str] = None

    @field_validator(
        "archiveRetentionTime",
        "dateOfDiskPurging",
        "dateOfPublishing",
        "publishedOn",
        mode="before",
    )
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @classmethod
    def user_model_type(cls) -> Type[Lifecycle]:
        return Lifecycle


class DownloadTechnique(BaseModel):
    name: Optional[str] = None
    pid: Optional[str] = None

    @classmethod
    def user_model_type(cls) -> Type[Technique]:
        return Technique

    @classmethod
    def upload_model_type(cls) -> Type[UploadTechnique]:
        return UploadTechnique


class UploadTechnique(BaseModel):
    name: str
    pid: str

    @classmethod
    def user_model_type(cls) -> Type[Technique]:
        return Technique

    @classmethod
    def download_model_type(cls) -> Type[DownloadTechnique]:
        return DownloadTechnique


class DownloadRelationship(BaseModel):
    pid: Optional[PID] = None
    relationship: Optional[str] = None

    @classmethod
    def user_model_type(cls) -> Type[Relationship]:
        return Relationship

    @classmethod
    def upload_model_type(cls) -> Type[UploadRelationship]:
        return UploadRelationship


class UploadRelationship(BaseModel):
    pid: PID
    relationship: str

    @classmethod
    def user_model_type(cls) -> Type[Relationship]:
        return Relationship

    @classmethod
    def download_model_type(cls) -> Type[DownloadRelationship]:
        return DownloadRelationship


class DownloadHistory(BaseModel):
    id: Optional[str] = pydantic.Field(alias="_id", default=None)
    updatedAt: Optional[datetime] = None
    updatedBy: Optional[datetime] = None

    @field_validator("updatedAt", mode="before")
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @classmethod
    def user_model_type(cls) -> Type[History]:
        return History


class DownloadDataFile(BaseModel):
    path: Optional[str] = None
    size: Optional[NonNegativeInt] = None
    time: Optional[datetime] = None
    chk: Optional[str] = None
    gid: Optional[str] = None
    perm: Optional[str] = None
    uid: Optional[str] = None

    @field_validator("time", mode="before")
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @classmethod
    def upload_model_type(cls) -> Type[UploadDataFile]:
        return UploadDataFile


class UploadDataFile(BaseModel):
    path: str
    size: NonNegativeInt
    time: datetime
    chk: Optional[str] = None
    gid: Optional[str] = None
    perm: Optional[str] = None
    uid: Optional[str] = None

    @field_validator("time", mode="before")
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @classmethod
    def download_model_type(cls) -> Type[DownloadDataFile]:
        return DownloadDataFile


class DownloadInstrument(BaseModel):
    customMetadata: Optional[Dict[str, Any]] = None
    name: Optional[str] = None
    pid: Optional[str] = None
    uniqueName: Optional[str] = None

    @classmethod
    def user_model_type(cls) -> Type[Instrument]:
        return Instrument


class DownloadSample(BaseModel):
    ownerGroup: Optional[str] = None
    accessGroups: Optional[List[str]] = None
    createdAt: Optional[datetime] = None
    createdBy: Optional[str] = None
    description: Optional[str] = None
    instrumentGroup: Optional[str] = None
    isPublished: Optional[bool] = None
    owner: Optional[str] = None
    sampleCharacteristics: Optional[Dict[str, Any]] = None
    sampleId: Optional[str] = None
    updatedAt: Optional[datetime] = None
    updatedBy: Optional[str] = None

    @field_validator("createdAt", "updatedAt", mode="before")
    def _validate_datetime(cls, value: Any) -> Any:
        return validate_datetime(value)

    @classmethod
    def user_model_type(cls) -> Type[Sample]:
        return Sample

    @classmethod
    def upload_model_type(cls) -> Type[UploadSample]:
        return UploadSample


class UploadSample(BaseModel):
    ownerGroup: str
    accessGroups: Optional[List[str]] = None
    description: Optional[str] = None
    instrumentGroup: Optional[str] = None
    isPublished: Optional[bool] = None
    owner: Optional[str] = None
    sampleCharacteristics: Optional[Dict[str, Any]] = None
    sampleId: Optional[str] = None

    @classmethod
    def user_model_type(cls) -> Type[Sample]:
        return Sample

    @classmethod
    def download_model_type(cls) -> Type[DownloadSample]:
        return DownloadSample


@dataclass_optional_args(kw_only=True, slots=True)
class Attachment(BaseUserModel):
    caption: str
    owner_group: str
    access_groups: Optional[List[str]] = None
    dataset_id: Optional[PID] = None
    id: Optional[str] = None
    instrument_group: Optional[str] = None
    proposal_id: Optional[str] = None
    sample_id: Optional[str] = None
    thumbnail: Optional[Thumbnail] = None
    _created_at: Optional[datetime] = None
    _created_by: Optional[str] = None
    _updated_at: Optional[datetime] = None
    _updated_by: Optional[str] = None

    @property
    def created_at(self) -> Optional[datetime]:
        return self._created_at

    @property
    def created_by(self) -> Optional[str]:
        return self._created_by

    @property
    def updated_at(self) -> Optional[datetime]:
        return self._updated_at

    @property
    def updated_by(self) -> Optional[str]:
        return self._updated_by

    @classmethod
    def from_download_model(cls, download_model: DownloadAttachment) -> Attachment:
        """Construct an instance from an associated SciCat download model."""
        return cls(**cls._download_model_dict(download_model))

    def make_upload_model(self) -> UploadAttachment:
        """Construct a SciCat upload model from self."""
        return UploadAttachment(**self._upload_model_dict())

    @classmethod
    def upload_model_type(cls) -> Type[UploadAttachment]:
        return UploadAttachment

    @classmethod
    def download_model_type(cls) -> Type[DownloadAttachment]:
        return DownloadAttachment


@dataclass_optional_args(kw_only=True, slots=True)
class Lifecycle(BaseUserModel):
    _archivable: Optional[bool] = None
    _archive_retention_time: Optional[datetime] = None
    _archive_return_message: Optional[Dict[str, Any]] = None
    _archive_status_message: Optional[str] = None
    _date_of_disk_purging: Optional[datetime] = None
    _date_of_publishing: Optional[datetime] = None
    _exported_to: Optional[str] = None
    _is_on_central_disk: Optional[bool] = None
    _publishable: Optional[bool] = None
    _published_on: Optional[datetime] = None
    _retrievable: Optional[bool] = None
    _retrieve_integrity_check: Optional[bool] = None
    _retrieve_return_message: Optional[Dict[str, Any]] = None
    _retrieve_status_message: Optional[str] = None

    @property
    def archivable(self) -> Optional[bool]:
        return self._archivable

    @property
    def archive_retention_time(self) -> Optional[datetime]:
        return self._archive_retention_time

    @property
    def archive_return_message(self) -> Optional[Dict[str, Any]]:
        return self._archive_return_message

    @property
    def archive_status_message(self) -> Optional[str]:
        return self._archive_status_message

    @property
    def date_of_disk_purging(self) -> Optional[datetime]:
        return self._date_of_disk_purging

    @property
    def date_of_publishing(self) -> Optional[datetime]:
        return self._date_of_publishing

    @property
    def exported_to(self) -> Optional[str]:
        return self._exported_to

    @property
    def is_on_central_disk(self) -> Optional[bool]:
        return self._is_on_central_disk

    @property
    def publishable(self) -> Optional[bool]:
        return self._publishable

    @property
    def published_on(self) -> Optional[datetime]:
        return self._published_on

    @property
    def retrievable(self) -> Optional[bool]:
        return self._retrievable

    @property
    def retrieve_integrity_check(self) -> Optional[bool]:
        return self._retrieve_integrity_check

    @property
    def retrieve_return_message(self) -> Optional[Dict[str, Any]]:
        return self._retrieve_return_message

    @property
    def retrieve_status_message(self) -> Optional[str]:
        return self._retrieve_status_message

    @classmethod
    def from_download_model(cls, download_model: DownloadLifecycle) -> Lifecycle:
        """Construct an instance from an associated SciCat download model."""
        return cls(**cls._download_model_dict(download_model))

    @classmethod
    def download_model_type(cls) -> Type[DownloadLifecycle]:
        return DownloadLifecycle


@dataclass_optional_args(kw_only=True, slots=True)
class Technique(BaseUserModel):
    name: str
    pid: str

    @classmethod
    def from_download_model(cls, download_model: DownloadTechnique) -> Technique:
        """Construct an instance from an associated SciCat download model."""
        return cls(**cls._download_model_dict(download_model))

    def make_upload_model(self) -> UploadTechnique:
        """Construct a SciCat upload model from self."""
        return UploadTechnique(**self._upload_model_dict())

    @classmethod
    def upload_model_type(cls) -> Type[UploadTechnique]:
        return UploadTechnique

    @classmethod
    def download_model_type(cls) -> Type[DownloadTechnique]:
        return DownloadTechnique


@dataclass_optional_args(kw_only=True, slots=True)
class Relationship(BaseUserModel):
    pid: PID
    relationship: str

    @classmethod
    def from_download_model(cls, download_model: DownloadRelationship) -> Relationship:
        """Construct an instance from an associated SciCat download model."""
        return cls(**cls._download_model_dict(download_model))

    def make_upload_model(self) -> UploadRelationship:
        """Construct a SciCat upload model from self."""
        return UploadRelationship(**self._upload_model_dict())

    @classmethod
    def upload_model_type(cls) -> Type[UploadRelationship]:
        return UploadRelationship

    @classmethod
    def download_model_type(cls) -> Type[DownloadRelationship]:
        return DownloadRelationship


@dataclass_optional_args(kw_only=True, slots=True)
class History(BaseUserModel):
    __id: Optional[str] = None
    _updated_at: Optional[datetime] = None
    _updated_by: Optional[datetime] = None

    @property
    def _id(self) -> Optional[str]:
        return self.__id

    @property
    def updated_at(self) -> Optional[datetime]:
        return self._updated_at

    @property
    def updated_by(self) -> Optional[datetime]:
        return self._updated_by

    @classmethod
    def from_download_model(cls, download_model: DownloadHistory) -> History:
        """Construct an instance from an associated SciCat download model."""
        return cls(**cls._download_model_dict(download_model))

    @classmethod
    def download_model_type(cls) -> Type[DownloadHistory]:
        return DownloadHistory


@dataclass_optional_args(kw_only=True, slots=True)
class Instrument(BaseUserModel):
    _custom_metadata: Optional[Dict[str, Any]] = None
    _name: Optional[str] = None
    _pid: Optional[str] = None
    _unique_name: Optional[str] = None

    @property
    def custom_metadata(self) -> Optional[Dict[str, Any]]:
        return self._custom_metadata

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def pid(self) -> Optional[str]:
        return self._pid

    @property
    def unique_name(self) -> Optional[str]:
        return self._unique_name

    @classmethod
    def from_download_model(cls, download_model: DownloadInstrument) -> Instrument:
        """Construct an instance from an associated SciCat download model."""
        return cls(**cls._download_model_dict(download_model))

    @classmethod
    def download_model_type(cls) -> Type[DownloadInstrument]:
        return DownloadInstrument


@dataclass_optional_args(kw_only=True, slots=True)
class Sample(BaseUserModel):
    owner_group: str
    access_groups: Optional[List[str]] = None
    description: Optional[str] = None
    instrument_group: Optional[str] = None
    is_published: Optional[bool] = None
    owner: Optional[str] = None
    sample_characteristics: Optional[Dict[str, Any]] = None
    sample_id: Optional[str] = None
    _created_at: Optional[datetime] = None
    _created_by: Optional[str] = None
    _updated_at: Optional[datetime] = None
    _updated_by: Optional[str] = None

    @property
    def created_at(self) -> Optional[datetime]:
        return self._created_at

    @property
    def created_by(self) -> Optional[str]:
        return self._created_by

    @property
    def updated_at(self) -> Optional[datetime]:
        return self._updated_at

    @property
    def updated_by(self) -> Optional[str]:
        return self._updated_by

    @classmethod
    def from_download_model(cls, download_model: DownloadSample) -> Sample:
        """Construct an instance from an associated SciCat download model."""
        return cls(**cls._download_model_dict(download_model))

    def make_upload_model(self) -> UploadSample:
        """Construct a SciCat upload model from self."""
        return UploadSample(**self._upload_model_dict())

    @classmethod
    def upload_model_type(cls) -> Type[UploadSample]:
        return UploadSample

    @classmethod
    def download_model_type(cls) -> Type[DownloadSample]:
        return DownloadSample


# Some models contain fields that are other models which are defined
# further down in the file.
# Instead of ordering models according to their dependencies, resolve
# references once all classes have been defined.
DownloadAttachment.model_rebuild()
UploadAttachment.model_rebuild()
DownloadOrigDatablock.model_rebuild()
UploadOrigDatablock.model_rebuild()
DownloadDatablock.model_rebuild()
UploadDatablock.model_rebuild()
DownloadLifecycle.model_rebuild()
DownloadTechnique.model_rebuild()
UploadTechnique.model_rebuild()
DownloadRelationship.model_rebuild()
UploadRelationship.model_rebuild()
DownloadHistory.model_rebuild()
DownloadDataFile.model_rebuild()
UploadDataFile.model_rebuild()
DownloadInstrument.model_rebuild()
DownloadSample.model_rebuild()
UploadSample.model_rebuild()
DownloadDataset.model_rebuild()
UploadDerivedDataset.model_rebuild()
UploadRawDataset.model_rebuild()

__all__ = (
    "BaseModel",
    "BaseUserModel",
    "DatasetType",
    "construct",
    "DownloadDataset",
    "UploadDerivedDataset",
    "UploadRawDataset",
    "DownloadAttachment",
    "UploadAttachment",
    "DownloadOrigDatablock",
    "UploadOrigDatablock",
    "DownloadDatablock",
    "UploadDatablock",
    "DownloadLifecycle",
    "DownloadTechnique",
    "UploadTechnique",
    "DownloadRelationship",
    "UploadRelationship",
    "DownloadHistory",
    "DownloadDataFile",
    "UploadDataFile",
    "DownloadInstrument",
    "DownloadSample",
    "UploadSample",
    "Attachment",
    "Lifecycle",
    "Technique",
    "Relationship",
    "History",
    "Instrument",
    "Sample",
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Handler for packaged assets."""

import importlib.resources
from functools import lru_cache
from string import Template


def _read_text(filename: str, group: str) -> str:
    return (
        importlib.resources.files(f"scitacean._html_repr.{group}")
        .joinpath(filename)
        .read_text()
    )


def _preprocess_style(css: str) -> str:
    import re

    # line breaks are not needed
    css = css.replace("\n", "")
    # remove comments
    css = re.sub(r"/\*(\*(?!/)|[^*])*\*/", "", css)
    # remove space around special characters
    css = re.sub(r"\s*([;{}:,])\s*", r"\1", css)
    return css


@lru_cache(maxsize=1)
def dataset_repr_template() -> Template:
    return Template(_read_text("dataset_repr.html.template", "templates"))


@lru_cache(maxsize=1)
def dataset_field_repr_template() -> Template:
    return Template(_read_text("dataset_field_repr.html.template", "templates"))


@lru_cache(maxsize=1)
def files_repr_template() -> Template:
    return Template(_read_text("files_repr.html.template", "templates"))


@lru_cache(maxsize=1)
def metadata_template() -> Template:
    return Template(_read_text("metadata_repr.html.template", "templates"))


@lru_cache(maxsize=1)
def attachment_repr_template() -> Template:
    return Template(_read_text("attachment_repr.html.template", "templates"))


@lru_cache(maxsize=1)
def attachment_field_repr_template() -> Template:
    return Template(_read_text("attachment_field_repr.html.template", "templates"))


@lru_cache(maxsize=1)
def common_style() -> str:
    sheet = _preprocess_style(_read_text("common.css", "styles"))
    return f"<style>{sheet}</style>"


@lru_cache(maxsize=1)
def dataset_style() -> str:
    sheet = _preprocess_style(_read_text("dataset.css", "styles"))
    return f"{common_style()}<style>{sheet}</style>"


@lru_cache(maxsize=1)
def attachment_style() -> str:
    sheet = _preprocess_style(_read_text("attachment.css", "styles"))
    return f"{common_style()}<style>{sheet}</style>"


@lru_cache()
def image(name: str) -> str:
    return _read_text(name, "images")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
import dataclasses
from typing import Type, TypeVar

import pytest
from dateutil.parser import parse as parse_date
from hypothesis import given, settings
from hypothesis import strategies as st

from scitacean import PID, DatasetType, RemotePath, model
from scitacean.model import (
    DownloadAttachment,
    DownloadDataset,
    DownloadOrigDatablock,
    UploadDerivedDataset,
    UploadRawDataset,
)

T = TypeVar("T")


def build_user_model_for_upload(cls: Type[T]) -> st.SearchStrategy[T]:
    private_fields = {
        field.name: st.none()
        for field in dataclasses.fields(cls)  # type: ignore[arg-type]
        if field.name.startswith("_")
    }
    return st.builds(cls, **private_fields)


@settings(max_examples=10)
@given(data=st.data())
@pytest.mark.parametrize(
    "model_types",
    (
        (model.Attachment, model.UploadAttachment),
        (model.Technique, model.UploadTechnique),
        (model.Relationship, model.UploadRelationship),
    ),
)
# Cannot test (model.Sample, model.UploadSample) because hypothesis
# cannot handle fields with type Any.
def test_can_make_upload_model(model_types, data):
    user_model_type, upload_model_type = model_types
    user_model = data.draw(build_user_model_for_upload(user_model_type))
    upload_model = user_model.make_upload_model()
    assert isinstance(upload_model, upload_model_type)


@settings(max_examples=10)
@given(build_user_model_for_upload(model.Attachment))
def test_upload_attachment_fields(attachment):
    upload_attachment = attachment.make_upload_model()
    assert upload_attachment.caption == attachment.caption
    assert upload_attachment.accessGroups == attachment.access_groups
    assert upload_attachment.thumbnail == attachment.thumbnail


@settings(max_examples=10)
@given(st.builds(model.Attachment))
def test_upload_model_rejects_non_upload_fields(attachment):
    attachment._created_by = "the-creator"
    with pytest.raises(ValueError):
        attachment.make_upload_model()


@settings(max_examples=10)
@given(data=st.data())
@pytest.mark.parametrize(
    "model_types",
    (
        (model.Attachment, model.DownloadAttachment),
        (model.Lifecycle, model.DownloadLifecycle),
        (model.Technique, model.DownloadTechnique),
        (model.History, model.DownloadHistory),
        (model.Instrument, model.DownloadInstrument),
        (model.Relationship, model.DownloadRelationship),
    ),
)
def test_can_make_from_download_model(model_types, data):
    user_model_type, download_model_type = model_types
    download_model = data.draw(st.builds(download_model_type))
    # doesn't raise
    user_model_type.from_download_model(download_model)


@settings(max_examples=10)
@given(st.builds(model.DownloadAttachment))
def test_download_attachment_fields(download_attachment):
    attachment = model.Attachment.from_download_model(download_attachment)
    assert attachment.caption == download_attachment.caption
    assert attachment.dataset_id == download_attachment.datasetId
    assert attachment.thumbnail == download_attachment.thumbnail


def test_derived_dataset_default_values(
    real_client, require_scicat_backend, scicat_access
):
    dset = UploadDerivedDataset(
        accessGroups=["access1"],
        contactEmail="contact@email.com",
        creationTime=parse_date("2000-01-01T01:01:01.000Z"),
        inputDatasets=[PID(prefix="PID.prefix.a0b1", pid="abcd")],
        investigator="inv@esti.gator",
        numberOfFilesArchived=0,
        owner=scicat_access.user.username,
        ownerGroup=scicat_access.user.group,
        sourceFolder=RemotePath("/source/folder"),
        usedSoftware=["software1"],
        type=DatasetType.DERIVED,
    )
    pid = real_client.scicat.create_dataset_model(dset).pid
    finalized = real_client.scicat.get_dataset_model(pid)

    # Inputs
    assert finalized.accessGroups == ["access1"]
    assert finalized.contactEmail == "contact@email.com"
    assert finalized.creationTime == parse_date("2000-01-01T01:01:01.000Z")
    assert finalized.inputDatasets == [PID(prefix="PID.prefix.a0b1", pid="abcd")]
    assert finalized.investigator == "inv@esti.gator"
    assert finalized.owner == scicat_access.user.username
    assert finalized.ownerGroup == scicat_access.user.group
    assert finalized.sourceFolder == "/source/folder"
    assert finalized.usedSoftware == ["software1"]

    # Default values
    assert finalized.createdAt  # some non-empty str
    assert finalized.createdBy  # some non-empty str
    assert finalized.classification  # some non-empty str
    assert finalized.datasetName  # some non-empty str
    assert finalized.history is None
    assert finalized.isPublished is False
    assert finalized.keywords == []
    assert finalized.numberOfFiles == 0
    assert finalized.numberOfFilesArchived == 0
    assert finalized.packedSize == 0
    assert finalized.pid  # some non-empty str
    assert finalized.scientificMetadata == {}
    assert finalized.sharedWith == []
    assert finalized.size == 0
    assert finalized.techniques == []
    assert finalized.updatedAt  # some non-empty str

    # Left empty
    assert finalized.description is None is None
    assert finalized.jobParameters is None
    assert finalized.jobLogData is None
    assert finalized.license is None
    assert finalized.orcidOfOwner is None
    assert finalized.ownerEmail is None
    assert finalized.sourceFolderHost is None
    assert finalized.validationStatus is None
    assert finalized.version is None


def test_raw_dataset_default_values(real_client, require_scicat_backend, scicat_access):
    dset = UploadRawDataset(
        accessGroups=["access1"],
        contactEmail="contact@email.com",
        creationTime=parse_date("2000-01-01T01:01:01.000Z"),
        creationLocation="site",
        numberOfFilesArchived=0,
        owner=scicat_access.user.username,
        ownerGroup=scicat_access.user.group,
        principalInvestigator="inv@esti.gator",
        sourceFolder=RemotePath("/source/folder"),
        type=DatasetType.RAW,
    )
    pid = real_client.scicat.create_dataset_model(dset).pid
    finalized = real_client.scicat.get_dataset_model(pid)

    # Inputs
    assert finalized.accessGroups == ["access1"]
    assert finalized.contactEmail == "contact@email.com"
    assert finalized.creationLocation == "site"
    assert finalized.creationTime == parse_date("2000-01-01T01:01:01.000Z")
    assert finalized.owner == scicat_access.user.username
    assert finalized.ownerGroup == scicat_access.user.group
    assert finalized.principalInvestigator == "inv@esti.gator"
    assert finalized.sourceFolder == "/source/folder"

    # Default values
    assert finalized.createdAt  # some non-empty str
    assert finalized.createdBy  # some non-empty str
    assert finalized.classification  # some non-empty str
    assert finalized.datasetName  # some non-empty str
    assert finalized.history is None
    assert finalized.isPublished is False
    assert finalized.keywords == []
    assert finalized.numberOfFiles == 0
    assert finalized.numberOfFilesArchived == 0
    assert finalized.packedSize == 0
    assert finalized.pid  # some non-empty str
    assert finalized.scientificMetadata == {}
    assert finalized.sharedWith == []
    assert finalized.size == 0
    assert finalized.techniques == []
    assert finalized.updatedAt  # some non-empty str

    # Left empty
    assert finalized.dataFormat is None
    assert finalized.description is None
    assert finalized.endTime is None
    assert finalized.instrumentId is None
    assert finalized.license is None
    assert finalized.orcidOfOwner is None
    assert finalized.ownerEmail is None
    assert finalized.proposalId is None
    assert finalized.sampleId is None
    assert finalized.sourceFolderHost is None
    assert finalized.validationStatus is None
    assert finalized.version is None


def test_default_masked_fields_are_dropped():
    mod = DownloadOrigDatablock(  # type: ignore[call-arg]
        id="abc",
        _v="123",
        __v="456",
    )
    # Input id dropped but alias 'id' exists and is not initialized.
    assert mod.id is None

    assert not hasattr(mod, "_v")
    assert not hasattr(mod, "__v")


def test_custom_masked_fields_are_dropped():
    mod = DownloadDataset(  # type: ignore[call-arg]
        attachments=[{"id": "abc"}],
        id="abc",
        _id="def",
        _v="123",
        __v="456",
    )
    assert not hasattr(mod, "attachments")
    assert not hasattr(mod, "id")
    assert not hasattr(mod, "_id")
    assert not hasattr(mod, "_v")
    assert not hasattr(mod, "__v")


def test_fields_override_masks():
    # '_id' is masked but the model has a field 'id' with alias '_id'.
    mod = DownloadOrigDatablock(  # type: ignore[call-arg]
        _id="abc",
        id="def",
    )
    assert mod.id == "abc"
    assert not hasattr(mod, "_id")


def test_fields_override_masks_att():
    # 'id' is masked but the model has a field 'id' without alias
    mod = DownloadAttachment(  # type: ignore[call-arg]
        _id="abc",
        id="def",
    )
    assert mod.id == "def"
    assert not hasattr(mod, "_id")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""Thumbnail type for encoding images."""

from __future__ import annotations

import base64
import mimetypes
import os
import re
from typing import Callable, Dict, Generator, Optional, Union

import pydantic

from ._internal.dataclass_wrapper import dataclass_optional_args
from ._internal.pydantic_compat import is_pydantic_v1

if not is_pydantic_v1():
    from typing import Any, Type

    from pydantic import GetCoreSchemaHandler
    from pydantic_core import core_schema

    def _get_thumbnail_core_schema(
        cls: Type[Thumbnail], _source_type: Any, _handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls.parse,
            core_schema.union_schema(
                [core_schema.is_instance_schema(Thumbnail), core_schema.str_schema()]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls.serialize, info_arg=False, return_schema=core_schema.str_schema()
            ),
        )


@dataclass_optional_args(init=False, kw_only=True, slots=True)
class Thumbnail:
    """Encodes an image to be used as a thumbnail in SciCat.

    Thumbnails are *small* images used, e.g., in attachments.
    In SciCat, they are base64-encoded strings and have a size limit.

    This class handles the encoding but does not enforce a size limit
    or content type.

    Currently, data is stored in encoded form.
    This means that creating a thumbnail object in a SciCat download is cheap.
    But creating one from local data and throwing it away without uploading it
    has some small overhead.

    Examples
    --------
    Given some raw bytes from a PNG, create a thumbnail using

    .. code-block:: python

        from scitacean import Thumbnail
        data = ... # the bytes of the PNG
        thumbnail = Thumbnail(mime="image/png", data=data)

    Or load the data directly from a file:

    .. code-block:: python

        thumbnail = Thumbnail.load_file("file_path.png")

    Access the raw bytes:

    .. code-block:: python

        data: bytes = thumbnail.decoded_data()
    """

    mime: Optional[str]
    """Complete MIME type in the form ``type/subtype``."""
    _encoded_data: str

    def __init__(
        self,
        mime: Optional[str],
        data: Optional[bytes] = None,
        _encoded_data: Optional[str] = None,
    ) -> None:
        """Create a new thumbnail object.

        Parameters
        ----------
        mime:
            The MIME type of the thumbnail.
            Must be a string of the form ``image/png``.
            Parameters are not allowed.
        data:
            The raw bytes of the thumbnail.
            Will be encoded automatically.
        _encoded_data:
            Primarily for internal use.
            Base64-encoded data of the thumbnail.
            Mutually exclusive with ``data``.
        """
        if data is None:
            if _encoded_data is None:
                raise TypeError("No thumbnail data specified")
            self._encoded_data = _encoded_data
        else:
            if _encoded_data is None:
                self._encoded_data = base64.b64encode(data).decode("utf-8")
            else:
                raise TypeError("Only only of data and _encoded_data may be given")

        self.mime = mime

    @classmethod
    def load_file(cls, path: Union[os.PathLike[str], str]) -> Thumbnail:
        """Construct a thumbnail from data loaded from a file.

        Parameters
        ----------
        path:
            The path to the file.

        Returns
        -------
        :
            A new thumbnail with MIME type guessed from the file
            and data loaded from disk.
        """
        with open(path, "rb") as f:
            data = f.read()
        encoded_data = base64.b64encode(data).decode("utf-8")
        return Thumbnail(mime=mimetypes.guess_type(path)[0], _encoded_data=encoded_data)

    @classmethod
    def parse(cls, encoded: Union[str, Thumbnail], /) -> Thumbnail:
        """Construct a thumbnail from a string as used by SciCat.

        Parameters
        ----------
        encoded:
            A string containing a MIME content-header and the thumbnail
            in base64 encoding.
            Or an existing ``Thumbnail`` instance which is copied on return.

        Returns
        -------
        :
            A new thumbnail with MIME type and data extracted from the string.

        See Also
        --------
        Thumbnail.serialize:
            The inverse operation.
        """
        if isinstance(encoded, Thumbnail):
            return Thumbnail(mime=encoded.mime, _encoded_data=encoded._encoded_data)

        if (match := _MESSAGE_REGEX.match(encoded)) is None:
            mime = None
            encoded_data = encoded
        else:
            mime = match[1]
            encoded_data = match[2]

        return Thumbnail(mime=mime, _encoded_data=encoded_data)

    def serialize(self) -> str:
        """Format the thumbnail into a string in the format expected by SciCat.

        Returns
        -------
        :
            A string containing the MIME content-header and the thumbnail
            in base64 encoding.

        See Also
        --------
        Thumbnail.parse:
            The inverse operation.
        """
        mime_str = f"data:{self.mime};base64," if self.mime is not None else ""
        return mime_str + self.encoded_data()

    @property
    def mime_type(self) -> Optional[str]:
        """The MIME type, i.e., the first part of ``type/subtype``."""
        if self.mime is None:
            return None
        return self.mime.split("/", 1)[0]

    @property
    def mime_subtype(self) -> Optional[str]:
        """The MIME subtype, i.e., the second part of ``type/subtype``."""
        if self.mime is None:
            return None
        return self.mime.split("/", 1)[1]

    def encoded_data(self) -> str:
        """Return the base64-encoded data of the thumbnail."""
        return self._encoded_data

    def decoded_data(self) -> bytes:
        """Return the raw bytes of the thumbnail."""
        return base64.b64decode(self._encoded_data)

    def __str__(self) -> str:
        return f"Thumbnail({self.mime}, len:{len(self.encoded_data())}B)"

    def __repr__(self) -> str:
        return f"Thumbnail(mime={self.mime}, data={self.decoded_data()!r})"

    def _repr_mimebundle_(
        self, include: Any = None, exclude: Any = None
    ) -> Dict[str, Union[bytes, str]]:
        def decoded() -> bytes:
            return self.decoded_data()

        repr_fns: dict[str, Callable[[], Union[bytes, str]]] = {
            "image/png": decoded,
            "image/jpeg": decoded,
            "image/svg+xml": decoded,
            "application/pdf": decoded,
            "text/html": lambda: f"<img src={self.serialize()}>",
            "text/plain": self.__str__,
        }
        if include is not None:
            repr_fns = {k: v for k, v in repr_fns.items() if k in include}
        if exclude is not None:
            repr_fns = {k: v for k, v in repr_fns.items() if k not in exclude}

        if self.mime in repr_fns:
            return {self.mime: repr_fns[self.mime]()}
        return {
            mime: fn()
            for mime in {"text/html", "text/plain"}
            if (fn := repr_fns.get(mime))
        }

    if is_pydantic_v1():

        @classmethod
        def validate(cls, value: Union[str, Thumbnail]) -> Thumbnail:
            """Pydantic validator for Thumbnail fields."""
            if isinstance(value, str):
                return Thumbnail.parse(value)
            if isinstance(value, Thumbnail):
                return value
            raise pydantic.ValidationError("expected a Thumbnail or str")

        @classmethod
        def __get_validators__(
            cls,
        ) -> Generator[Callable[[Union[str, Thumbnail]], Thumbnail], None, None]:
            yield cls.validate

    else:
        __get_pydantic_core_schema__ = classmethod(_get_thumbnail_core_schema)


# A regex that matches encoded thumbnails with header.
# Expected form: data:image/png;base64,the-data
_MESSAGE_REGEX = re.compile(
    r"^(?:data:)?"  # optional data: prefix
    "(?:([^/]+/[^;]+)(?:;.*)?,)?"  # MIME content-header
    "(.*)$"  # data
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""HTML representations for attachments for Jupyter."""

from __future__ import annotations

import dataclasses
from typing import List

from ..model import Attachment, UploadAttachment
from . import _resources
from ._common_html import Field, format_field_flag, format_type, format_value


def attachment_html_repr(attachment: Attachment) -> str:
    template = _resources.attachment_repr_template()
    style_sheet = _resources.attachment_style()
    rows = "\n".join(_format_field(field) for field in _get_fields(attachment))
    if attachment.thumbnail is None:
        thumbnail = ""
    else:
        bundle = attachment.thumbnail._repr_mimebundle_(include=("text/html",))
        thumbnail = bundle["text/html"]  # type: ignore[assignment]
    return template.substitute(
        style_sheet=style_sheet,
        caption=format_value(attachment.caption),
        rows=rows,
        thumbnail=thumbnail,
    )


def _format_field(field: Field) -> str:
    name = field.name
    flag = format_field_flag(field)
    typ = format_type(field.type)
    value = format_value(field.value)

    template = _resources.attachment_field_repr_template()
    return template.substitute(
        name=name,
        flag=flag,
        type=typ,
        value=value,
    )


_EXCLUDED_FIELDS = {
    "caption",
    "thumbnail",
}


def _get_fields(attachment: Attachment) -> List[Field]:
    fields = [
        Field(
            name=_strip_leading_underscore(field.name),
            value=getattr(attachment, field.name),
            type=field.type,
            description="",
            read_only=_is_read_only(field.name),
            required=False,
            error=None,
            main=True,
        )
        for field in dataclasses.fields(attachment)
        if field.name not in _EXCLUDED_FIELDS
    ]
    return sorted(fields, key=lambda field: field.name)


def _strip_leading_underscore(s: str) -> str:
    return s[1:] if s.startswith("_") else s


def _is_read_only(field_name: str) -> bool:
    return field_name not in UploadAttachment.get_model_fields()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 SciCat Project (https://github.com/SciCatProject/scitacean)
"""File transfers to upload and download files to / from a file server.

All file transfers are designed to work with :class:`scitacean.Client`.
Refer to the submodules for concrete setup and usage instructions.
"""
